# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Common tools for exogenous HDF5 format support
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import abc
import os.path as osp
from collections.abc import Callable

import h5py
import numpy as np

from cdl.config import Conf
from cdl.core.io.conv import convert_array_to_standard_type, data_to_xy
from cdl.utils.strings import to_string


class BaseNode(metaclass=abc.ABCMeta):
    """Object representing a HDF5 node"""

    IS_ARRAY = False

    def __init__(self, h5file, dname):
        self.h5file = h5file
        self.dset = h5file[dname]
        self.metadata = {}
        self.__obj = None
        self.children = []
        self.uint32_wng = False

    @property
    def id(self):
        """Return node id"""
        return self.dset.name

    @property
    def name(self):
        """Return node name, constructed from dataset name"""
        return to_string(self.dset.name).split("/")[-1]

    @property
    def data(self):
        """Data associated to node, if available"""
        return None

    @property
    def icon_name(self):
        """Icon name associated to node"""

    @property
    def shape_str(self):
        """Return string representation of node shape, if any"""
        return ""

    @property
    def dtype_str(self):
        """Return string representation of node data type, if any"""
        return ""

    @property
    def text(self):
        """Return node textual representation"""
        return ""

    @property
    def description(self):
        """Return node description"""
        return ""

    @classmethod
    def match(cls, dset):
        """Return True if h5 dataset match node pattern"""

    def is_supported(self) -> bool:
        """Return True if node is associated to supported data"""
        return False

    def create_native_object(self):
        """Create native object, if supported"""
        return None

    def get_native_object(self):
        """Return native object, if supported"""
        if self.__obj is None:
            obj = self.create_native_object()  # pylint: disable=assignment-from-none
            if obj is not None:
                self.__process_metadata(obj)
            self.__obj = obj
        return self.__obj

    def collect_attributes(self):
        """Collect attributes from node"""
        for key, value in self.dset.attrs.items():
            if isinstance(value, bytes):
                value = to_string(value)
            if isinstance(value, (np.ndarray, str, float, int, bool)) or np.isscalar(
                value
            ):
                self.metadata[key] = value

    def __process_metadata(self, obj):
        """Process metadata from dataset to obj"""
        obj.reset_metadata_to_defaults()
        obj.metadata["HDF5Path"] = self.h5file.filename
        obj.metadata["HDF5Dataset"] = self.id
        obj.metadata.update(self.metadata)

    @property
    def object_title(self):
        """Return signal/image object title"""
        if Conf.io.h5_fullpath_in_title.get():
            title = self.id
        else:
            title = self.name
        if Conf.io.h5_fname_in_title.get():
            title += f" ({osp.basename(self.h5file.filename)})"
        return title

    def set_signal_data(self, obj):
        """Set signal data (handles various issues)"""
        data = self.data
        if data.dtype not in (float, np.complex128):
            data = np.array(data, dtype=float)
        data = convert_array_to_standard_type(data)
        if len(data.shape) == 1:
            obj.set_xydata(np.arange(data.size), data)
        else:
            x, y, dx, dy = data_to_xy(data)
            obj.set_xydata(x, y, dx, dy)

    def set_image_data(self, obj):
        """Set image data (handles various issues)"""
        data = self.data
        if data.dtype == np.uint32:
            self.uint32_wng = data.max() > np.iinfo(np.int32).max
            clipped_data = data.clip(0, np.iinfo(np.int32).max)
            data = np.array(clipped_data, dtype=np.int32)
        obj.data = convert_array_to_standard_type(data)


class H5Importer:
    """DataLab HDF5 importer class"""

    def __init__(self, filename):
        self.h5file = h5py.File(filename)
        self.__nodes = {}
        self.root = RootNode(self.h5file)
        self.__nodes[self.root.id] = self.root
        self.root.collect_children(self.__nodes)
        NODE_FACTORY.run_post_triggers(self)

    @property
    def nodes(self):
        """Return all nodes"""
        return self.__nodes.values()

    def get(self, node_id: str):
        """Return node associated to id"""
        return self.__nodes[node_id]

    def get_relative(self, node: BaseNode, relpath: str, ancestor: int = 0):
        """Return node using relative path to another node"""
        path = "/" + (
            "/".join(node.id.split("/")[:-ancestor]) + "/" + relpath.strip("/")
        ).strip("/")
        return self.__nodes[path]

    def close(self):
        """Close HDF5 file"""
        self.__nodes = {}
        self.h5file.close()


class NodeFactory:
    """Factory for node classes"""

    def __init__(self):
        self.__ignored_datasets = []
        self.__generic_classes = []
        self.__thirdparty_classes = []
        self.__post_triggers = {}

    def add_ignored_datasets(self, names):
        """Add h5 dataset name to ignore list"""
        self.__ignored_datasets.extend(names)

    def add_post_trigger(self, nodecls: BaseNode, callback: Callable):
        """Add post trigger function, to be called at the end of the collect process.
        Callbacks take only one argument: H5Importer instance."""
        triggers = self.__post_triggers.setdefault(nodecls, [])
        triggers.append(callback)

    def register(self, cls, is_generic=False):
        """Register node class.
        Generic classes are processed after specific classes (as a fallback solution)"""
        if is_generic:
            self.__generic_classes.append(cls)
        else:
            self.__thirdparty_classes.append(cls)

    def get(self, dset):
        """Return node class that matches h5 dataset"""
        for name in to_string(dset.name).split("/"):
            if name in self.__ignored_datasets:
                return None
        for cls in self.__thirdparty_classes + self.__generic_classes:
            if cls.match(dset):
                return cls
        if isinstance(dset, h5py.Group):
            return GroupNode
        return None

    def run_post_triggers(self, importer: H5Importer):
        """Run post-collect callbacks"""
        for node in importer.nodes:
            for nodecls, triggers in self.__post_triggers.items():
                if isinstance(node, nodecls):
                    for func in triggers:
                        func(node, importer)


NODE_FACTORY = NodeFactory()


class GroupNode(BaseNode):
    """Object representing a HDF5 group node"""

    @property
    def icon_name(self):
        """Icon name associated to node"""
        return "h5group.svg"

    def collect_children(self, node_dict: dict[str, BaseNode]):
        """Construct tree"""
        for dset in self.dset.values():
            child_cls = NODE_FACTORY.get(dset)
            if child_cls is not None:
                child = child_cls(self.h5file, dset.name)
                node_dict[child.id] = child
                self.children.append(child)
                if isinstance(child, GroupNode):
                    child.collect_children(node_dict)
                child.collect_attributes()


class RootNode(GroupNode):
    """Object representing a HDF5 root node"""

    def __init__(self, h5file):
        super().__init__(h5file, "/")

    @property
    def icon_name(self):
        """Icon name associated to node"""
        return "h5file.svg"

    @property
    def name(self):
        """Return node name, constructed from dataset name"""
        return osp.basename(self.h5file.filename)

    @property
    def description(self):
        """Return node description"""
        return self.h5file.filename



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Process isolation unit test
---------------------------
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: skip

import time
from multiprocessing import Pool

import numpy as np
import scipy.signal as sps
from guidata.configtools import get_icon
from guidata.qthelpers import qt_app_context
from plotpy.plot import PlotOptions, PlotWidget
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

import cdl.obj
from cdl.env import execenv
from cdl.tests.data import create_2d_random, create_noisygauss_image


class MainWindow(QW.QMainWindow):
    """Multiprocessing test main window"""

    SIG_COMPUTATION_FINISHED = QC.Signal()

    def __init__(self):
        super().__init__()

        self.setWindowTitle("DataLab Multiprocessing test")
        self.setWindowIcon(get_icon("datalab.svg"))

        # Setting up the layout and widgets
        self.imagewidget = PlotWidget(options=PlotOptions(type="image"))
        self.setCentralWidget(self.imagewidget)
        self.layout = QW.QVBoxLayout(self.imagewidget)

        self.start_button = QW.QPushButton("Start Computation", self)
        self.start_button.clicked.connect(self.start_computation)
        self.layout.addWidget(self.start_button)

        self.cancel_button = QW.QPushButton("Cancel Computation", self)
        self.cancel_button.clicked.connect(self.cancel_computation)
        self.layout.addWidget(self.cancel_button)

        # Create a test image and add it to the plot
        param = cdl.obj.new_image_param(height=1000, width=1000)
        image = create_noisygauss_image(param, add_annotations=True)
        self.imageitem = image.make_item()
        self.imagewidget.plot.add_item(self.imageitem)

        self.array = image.data
        self.result = None
        self.timer = QC.QTimer()
        self.timer.setInterval(10)  # Check every 100 ms
        self.timer.timeout.connect(self.check_process)
        self.start_time = None

        self.SIG_COMPUTATION_FINISHED.connect(self.update_plot)

    @staticmethod
    def long_running_task(array: np.ndarray) -> None:
        """
        A long running task that computes a median filter on the input array and puts
        the result in the queue. The done_event is set when the computation is finished.

        Args:
            array (numpy.ndarray): The input data to compute the filter on.
        """
        start_time = time.time()
        result = sps.medfilt(array, 1) + create_2d_random(array.shape[0], array.dtype)
        execenv.print(f"Computation done: delta={time.time() - start_time:.3f} s")
        return result

    def print_time(self, title: str) -> None:
        """
        Prints the time since the last call to this method and resets the start_time.

        Args:
            title (str): The title to print before the time.
        """
        execenv.print(f"{title}: {time.time() - self.start_time:.3f} s")
        self.start_time = time.time()

    def start_computation(self) -> None:
        """
        Starts the computation in a separate process and starts the timer.
        """
        global POOL  # pylint: disable=global-statement,global-variable-not-assigned
        if self.result and not self.result.ready():
            self.print_time("Computation already running!")
            return
        self.start_time = time.time()
        self.result = POOL.apply_async(self.long_running_task, (self.array,))
        self.timer.start()
        self.print_time("Computation started")

    def cancel_computation(self) -> None:
        """
        Cancels the computation by terminating the process and stopping the timer.
        """
        global POOL  # pylint: disable=global-statement
        # Terminate the process and stop the timer
        POOL.terminate()
        POOL.join()
        self.timer.stop()
        execenv.print("Computation cancelled!")
        # Recreate the pool for the next computation
        POOL = Pool(processes=1)  # pylint: disable=not-callable,consider-using-with

    def check_process(self) -> None:
        """
        Checks if the computation is finished. If it is, stops the timer and calls
        the method to handle the computation finishing.
        """
        if self.result.ready():
            self.print_time("Computation finished")
            self.timer.stop()
            self.on_computation_finished()

    def on_computation_finished(self) -> None:
        """
        Retrieves the result from the queue
        and emits the SIG_COMPUTATION_FINISHED signal.
        """
        self.array = self.result.get()
        self.SIG_COMPUTATION_FINISHED.emit()
        self.print_time("Computation result retrieved")

    def update_plot(self) -> None:
        """
        Updates the plot with the result of the computation.
        """
        self.imageitem.set_data(self.array)
        self.imagewidget.plot.replot()
        self.print_time("Plot updated")

    def closeEvent(self, event) -> None:
        """
        Overrides the closeEvent to stop the timer
        and terminate the process when the window is closed.

        Args:
            event: The close event.
        """
        self.timer.stop()
        super().closeEvent(event)


def test_multiprocessing2() -> None:
    """
    Creates a PyQt application context, shows the main window,
    and starts the event loop.
    """
    global POOL  # pylint: disable=global-statement,global-variable-not-assigned
    with qt_app_context(exec_loop=True):
        window = MainWindow()
        window.show()
    POOL.terminate()
    POOL.join()


if __name__ == "__main__":
    POOL = Pool(processes=1)  # pylint: disable=not-callable,consider-using-with
    test_multiprocessing2()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Remote client test
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: skip

import os
import os.path as osp
import time

import numpy as np
from guidata.qthelpers import qt_app_context
from plotpy.builder import make

from cdl import app
from cdl.env import execenv
from cdl.param import XYCalibrateParam
from cdl.proxy import RemoteProxy
from cdl.tests.data import create_2d_gaussian, create_paracetamol_signal
from cdl.utils.tests import CDLTemporaryDirectory, exec_script


def multiple_commands(remote: RemoteProxy):
    """Execute multiple XML-RPC commands"""
    with CDLTemporaryDirectory() as tmpdir:
        x, y = create_paracetamol_signal().get_data()
        remote.add_signal("tutu", x, y)

        z = create_2d_gaussian(2000, np.uint16)
        remote.add_image("toto", z)
        rect = make.annotated_rectangle(100, 100, 200, 200, title="Test")
        area = rect.get_rect()
        remote.add_annotations_from_items([rect])
        uuid = remote.get_sel_object_uuids()[0]
        items = remote.get_object_shapes()
        assert len(items) == 1 and items[0].get_rect() == area
        remote.add_label_with_title(f"Image uuid: {uuid}")
        remote.select_groups([1])
        remote.select_objects([uuid])
        remote.delete_metadata()

        fname = osp.join(tmpdir, osp.basename("remote_test.h5"))
        remote.save_to_h5_file(fname)
        remote.reset_all()
        remote.open_h5_files([fname], True, False)
        remote.import_h5_file(fname, True)
        remote.set_current_panel("signal")
        assert remote.get_current_panel() == "signal"
        remote.calc("log10")

        param = XYCalibrateParam.create(a=1.2, b=0.1)
        remote.calc("compute_calibration", param)

        time.sleep(2)  # Avoid permission error when trying to clean-up temporary files


def test_remoteclient_unit():
    """Remote client test"""
    env = os.environ.copy()
    env[execenv.DO_NOT_QUIT_ENV] = "1"
    execenv.print("Launching DataLab in a separate process")
    exec_script(app.__file__, wait=False, env=env)
    remote = RemoteProxy()
    execenv.print("Executing multiple commands...", end="")
    with qt_app_context():  # needed for building plot items
        multiple_commands(remote)
    execenv.print("OK")


if __name__ == "__main__":
    test_remoteclient_unit()



"""
Computation (:mod:`cdl.computation`)
-----------------------------------------

This package contains the computation functions used by the DataLab project.
Those functions operate directly on DataLab objects (i.e. :class:`cdl.obj.SignalObj`
and :class:`cdl.obj.ImageObj`) and are designed to be used in the DataLab pipeline,
but can be used independently as well.

.. seealso::

    The :mod:`cdl.computation` package is the main entry point for the DataLab
    computation functions when manipulating DataLab objects.
    See the :mod:`cdl.algorithms` package for algorithms that operate directly on
    NumPy arrays.

Each computation module defines a set of computation objects, that is, functions
that implement processing features and classes that implement the corresponding
parameters (in the form of :py:class:`guidata.dataset.datatypes.Dataset` subclasses).
The computation functions takes a DataLab object (e.g. :class:`cdl.obj.SignalObj`)
and a parameter object (e.g. :py:class:`cdl.param.MovingAverageParam`) as input
and return a DataLab object as output (the result of the computation). The parameter
object is used to configure the computation function (e.g. the size of the moving
average window).

In DataLab overall architecture, the purpose of this package is to provide the
computation functions that are used by the :mod:`cdl.core.gui.processor` module,
based on the algorithms defined in the :mod:`cdl.algorithms` module and on the
data model defined in the :mod:`cdl.obj` (or :mod:`cdl.core.model`) module.

The computation modules are organized in subpackages according to their purpose.
The following subpackages are available:

- :mod:`cdl.computation.base`: Common processing features
- :mod:`cdl.computation.signal`: Signal processing features
- :mod:`cdl.computation.image`: Image processing features

Common processing features
^^^^^^^^^^^^^^^^^^^^^^^^^^

.. automodule:: cdl.computation.base
   :members:

Signal processing features
^^^^^^^^^^^^^^^^^^^^^^^^^^

.. automodule:: cdl.computation.signal
   :members:

Image processing features
^^^^^^^^^^^^^^^^^^^^^^^^^

Base image processing features
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. automodule:: cdl.computation.image
   :members:

Threshold features
~~~~~~~~~~~~~~~~~~

.. automodule:: cdl.computation.image.threshold
    :members:

Exposure correction features
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. automodule:: cdl.computation.image.exposure
    :members:

Restoration features
~~~~~~~~~~~~~~~~~~~~

.. automodule:: cdl.computation.image.restoration
    :members:

Morphological features
~~~~~~~~~~~~~~~~~~~~~~

.. automodule:: cdl.computation.image.morphology
    :members:

Edge detection features
~~~~~~~~~~~~~~~~~~~~~~~

.. automodule:: cdl.computation.image.edges

Detection features
~~~~~~~~~~~~~~~~~~

.. automodule:: cdl.computation.image.detection
    :members:
"""



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
GDI objects loading test
========================

This test is specific to Windows. It is not relevant for other platforms.

This test aims at checking that all GDI objects are released when the
widget are closed (i.e. when the Python object is garbage collected).
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

# guitest: skip

from __future__ import annotations

import ctypes
import os

import cdl.obj
from cdl.core.gui.main import CDLMainWindow
from cdl.env import execenv
from cdl.tests import cdltest_app_context
from cdl.tests.data import create_sincos_image
from cdl.tests.features.common.newobject_unit_test import iterate_image_creation
from cdl.tests.scenarios.common import compute_common_operations
from cdl.utils.tests import get_test_fnames

if os.name == "nt":
    from ctypes import WinDLL

    def get_gdi_count() -> int:
        """Get the number of GDI objects for the current process.

        This function uses the Windows API to get the count of GDI (Graphical
        Device Interface) objects used by the current process.

        Note: This function will only work on Windows.

        Returns:
            int: The count of GDI objects for the current process.
        """
        # Constants
        GR_GDIOBJECTS: int = 0

        # Load the User32 DLL
        user32: WinDLL = ctypes.windll.user32

        # Get the current process id
        pid: int = os.getpid()

        # Get handle of the process
        handle: int = ctypes.windll.kernel32.OpenProcess(1040, 0, pid)

        # Call the function and get the GDI count
        gdi_count: int = user32.GetGuiResources(handle, GR_GDIOBJECTS)

        # Close the handle
        ctypes.windll.kernel32.CloseHandle(handle)

        return gdi_count

else:

    def get_gdi_count() -> int:
        """Dumb function that always returns 0."""
        return 0


def test_various_image_features(win: CDLMainWindow):
    """Run image related tests."""
    win.set_current_panel("image")
    panel = win.imagepanel
    param = cdl.obj.new_image_param(height=150, width=150)
    for image in iterate_image_creation(param.width, non_zero=True, verbose=False):
        panel.add_object(create_sincos_image(param))
        panel.add_object(image)
        compute_common_operations(panel)
        panel.remove_all_objects()
        break


def test_gdi_count(win: CDLMainWindow) -> int | None:
    """Test the GDI count.

    This function will create a DataSetGroup, show the widget, and then
    close the widget. It will then check that the GDI count has not
    increased.

    Raises:
        AssertionError: If the GDI count has increased.

    Returns:
        int: The GDI count after creating the widget.
    """
    # Get the GDI count before creating the widget
    gdi_count_before: int = get_gdi_count()

    execenv.print(f"   GDI count: {gdi_count_before} --> ", end="")

    # Create widgets during the test
    test_various_image_features(win)

    # Import HDF5 file using the HDF5 browser
    win.open_h5_files(get_test_fnames("*.h5")[:5], import_all=True, reset_all=False)
    for panel in (win.signalpanel, win.imagepanel):
        panel.remove_all_objects()

    # Get the GDI count after creating the widget
    gdi_count_after: int = get_gdi_count()

    execenv.print(gdi_count_after)

    # # Check that the GDI count has not increased
    # assert gdi_count_before == gdi_count_after
    # assert gdi_count_before == gdi_count_after_close

    return gdi_count_after


def load_test():
    """Load test."""
    with execenv.context(unattended=True):
        with cdltest_app_context() as win:
            gdi_count = []
            for iteration in range(4):
                execenv.print(f"Test iteration: {iteration}")
                count = test_gdi_count(win)
                if count is None:
                    execenv.print("Test aborted")
                    break
                gdi_count.append(count)
                if iteration > 0:
                    increase = gdi_count[-1] - gdi_count[-2]
                    assert increase <= 0, "GDI count should not increase (memory leak)"
                    increase_pct = increase / gdi_count[0] * 100
                    execenv.print(
                        f"   GDI count increase: {increase:+d} ({increase_pct:.2f}%)"
                    )


if __name__ == "__main__":
    load_test()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab remote controlling class for Python 2.7
"""

import io
import os
import os.path as osp
import socket
import sys

import ConfigParser as cp
import numpy as np
from guidata.userconfig import get_config_dir
from xmlrpclib import Binary, ServerProxy


def array_to_rpcbinary(data):
    """Convert NumPy array to XML-RPC Binary object, with shape and dtype"""
    dbytes = io.BytesIO()
    np.save(dbytes, data, allow_pickle=False)
    return Binary(dbytes.getvalue())


def get_cdl_xmlrpc_port():
    """Return DataLab current XML-RPC port"""
    if sys.platform == "win32" and "HOME" in os.environ:
        os.environ.pop("HOME")  # Avoid getting old WinPython settings dir
    fname = osp.join(get_config_dir(), ".DataLab", "DataLab.ini")
    ini = cp.ConfigParser()
    ini.read(fname)
    try:
        return ini.get("main", "rpc_server_port")
    except (cp.NoSectionError, cp.NoOptionError):
        raise ConnectionRefusedError("DataLab has not yet been executed")


class RemoteClient(object):
    """Object representing a proxy/client to DataLab XML-RPC server"""

    def __init__(self):
        self.port = None
        self.serverproxy = None

    def connect(self, port=None):
        """Connect to DataLab XML-RPC server"""
        if port is None:
            port = get_cdl_xmlrpc_port()
        self.port = port
        url = "http://127.0.0.1:" + port
        self.serverproxy = ServerProxy(url, allow_none=True)
        try:
            self.get_version()
        except socket.error:
            raise ConnectionRefusedError("DataLab is currently not running")

    def get_version(self):
        """Return DataLab public version"""
        return self.serverproxy.get_version()

    def close_application(self):
        """Close DataLab application"""
        self.serverproxy.close_application()

    def raise_window(self):
        """Raise DataLab window"""
        self.serverproxy.raise_window()

    def get_current_panel(self):
        """Return current panel"""
        return self.serverproxy.get_current_panel()

    def set_current_panel(self, panel):
        """Switch to panel"""
        self.serverproxy.set_current_panel(panel)

    def reset_all(self):
        """Reset all application data"""
        self.serverproxy.reset_all()

    def toggle_auto_refresh(self, state):
        """Toggle auto refresh state"""
        self.serverproxy.toggle_auto_refresh(state)

    def toggle_show_titles(self, state):
        """Toggle show titles state"""
        self.serverproxy.toggle_show_titles(state)

    def save_to_h5_file(self, filename):
        """Save to a DataLab HDF5 file"""
        self.serverproxy.save_to_h5_file(filename)

    def open_h5_files(self, h5files, import_all, reset_all):
        """Open a DataLab HDF5 file or import from any other HDF5 file"""
        self.serverproxy.open_h5_files(h5files, import_all, reset_all)

    def import_h5_file(self, filename, reset_all):
        """Open DataLab HDF5 browser to Import HDF5 file"""
        self.serverproxy.import_h5_file(filename, reset_all)

    def load_from_files(self, filenames):
        """Open objects from files in current panel (signals/images)"""
        self.serverproxy.load_from_files(filenames)

    def add_signal(
        self, title, xdata, ydata, xunit=None, yunit=None, xlabel=None, ylabel=None
    ):
        """Add signal data to DataLab"""
        xbinary = array_to_rpcbinary(xdata)
        ybinary = array_to_rpcbinary(ydata)
        p = self.serverproxy
        return p.add_signal(title, xbinary, ybinary, xunit, yunit, xlabel, ylabel)

    def add_image(
        self,
        title,
        data,
        xunit=None,
        yunit=None,
        zunit=None,
        xlabel=None,
        ylabel=None,
        zlabel=None,
    ):
        """Add image data to DataLab"""
        zbinary = array_to_rpcbinary(data)
        p = self.serverproxy
        return p.add_image(title, zbinary, xunit, yunit, zunit, xlabel, ylabel, zlabel)

    def get_object_titles(self, panel=None):
        """Get object (signal/image) list for current panel"""
        return self.serverproxy.get_object_titles(panel)

    def get_object(self, nb_id_title=None, panel=None):
        """Get object (signal/image) by number, id or title"""
        return self.serverproxy.get_object(nb_id_title, panel)

    def get_object_uuids(self, panel=None):
        """Get object (signal/image) list for current panel"""
        return self.serverproxy.get_object_uuids(panel)


def test_remote_client():
    """DataLab Remote Client test"""
    cdl = RemoteClient()
    cdl.connect()
    data = np.array([[3, 4, 5], [7, 8, 0]], dtype=np.uint16)
    cdl.add_image("toto", data)


if __name__ == "__main__":
    test_remote_client()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Load application test: high number of images
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from __future__ import annotations

import numpy as np
import pytest

import cdl.obj
import cdl.param as dlp
from cdl.env import execenv
from cdl.tests import cdltest_app_context


def create_random_test_data(size: tuple[int, int] | None = None) -> cdl.obj.ImageObj:
    """Create a test image, based on a fast algorithm, to be able to generate
    a high number of images.

    Args:
        size: Size of the image.
    """
    if size is None:
        size = (2048, 2048)
    # Create a base image with low frequency shapes:
    x = np.linspace(-1, 1, size[0])
    y = np.linspace(-1, 1, size[1])
    xx, yy = np.meshgrid(x, y)
    a1, a2, f1, f2 = np.random.rand(4)
    data = a1 * np.sin(2 * np.pi * xx * (1 + f1 * 0.2)) + a2 * np.cos(
        2 * np.pi * yy * (1 + f2 * 0.2)
    )
    # Add some random noise:
    data += 0.1 * np.random.randn(*data.shape)
    image = cdl.obj.create_image("Random test image", data)
    return image


@pytest.mark.skip("This a load test, not a functional test")
def test_high_number_of_images() -> None:
    """Run a test with a high number of images."""
    nb = 30
    execenv.print("Creating images", end="")
    images = []
    for idx in range(nb):
        execenv.print(".", end="")
        ima = create_random_test_data()
        ima.title += f" {idx}"
        images.append(ima)
    execenv.print(" done")
    with cdltest_app_context() as win:
        panel = win.imagepanel
        for ima in images:
            panel.add_object(ima)
        panel.objview.select_groups()

        # Comment the two following lines to check if DataLab only shows the last image
        # when they are all superposed
        param = dlp.GridParam.create(cols=10)
        panel.processor.distribute_on_grid(param)

        panel.duplicate_object()


if __name__ == "__main__":
    test_high_number_of_images()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Operation modes test
--------------------

DataLab has two operation modes:

- **Single operand mode**: the operation is applied to the selected objects (this
  is the default mode)

- **Pairwise mode**: the operation is applied to the selected pairs of objects

This test scenario covers the pairwise mode and the operations that can be
performed in this mode: sum, difference, product, division, ...
"""

# guitest: show

from __future__ import annotations

from cdl import app
from cdl.config import Conf
from cdl.core.gui.processor.base import is_pairwise_mode
from cdl.env import execenv
from cdl.utils.qthelpers import cdl_app_context
from cdl.utils.tests import get_test_fnames


def check_titles(title, titles):
    """Check that the title is one of the expected titles"""
    execenv.print(f"{title}:")
    for actual_title, expected_title in titles:
        execenv.print(f"  {actual_title} == {expected_title}", end=" ")
        assert actual_title == expected_title
        if actual_title == expected_title:
            execenv.print("✓")
        else:
            execenv.print("✗")


def test_single_operand_mode_compute_n1():
    """Run single operand mode test scenario
    with compute_n1 operation (e.g. sum)"""
    original_mode = Conf.proc.operation_mode.get()
    Conf.proc.operation_mode.set("single")

    with cdl_app_context(exec_loop=True):
        win = app.create(h5files=[get_test_fnames("reorder*")[0]], console=False)
        panel = win.signalpanel
        view, model = panel.objview, panel.objmodel

        # Checking the operation mode:
        assert not is_pairwise_mode()

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select the two first groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        view.select_groups(groups)

        # Perform a sum operation
        panel.processor.compute_sum()

        # Default operation mode is single operand mode, so the sum operation
        # is applied to the selected groups, and we should have a new group
        # with two signals being the sum of signals from each group:
        # - signal 1: group 1 signal 1 + group 1 signal 2
        # - signal 2: group 2 signal 1 + group 2 signal 2
        assert len(model.get_groups()) == n_groups + 1
        new_group = model.get_group_from_number(n_groups + 1)
        assert len(new_group) == 2
        titles = []
        for idx, obj in enumerate(new_group):
            pfx_orig = ", ".join(obj.short_id for obj in groups[idx].get_objects())
            titles.append((obj.title, f"Σ({pfx_orig})"))
        check_titles(f"Single operand mode Σ[{new_group.title}]", titles)

        # Remove new group
        view.select_groups([new_group])
        panel.remove_object(force=True)

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select the two first signals of the first two groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        objs = groups[0][:2] + groups[1][:2]
        view.select_objects(objs)

        # Perform a sum operation
        panel.processor.compute_sum()

        # Default operation mode is single operand mode, so the sum operation
        # is applied to the selected signals, and we should have a new resulting
        # signal being the sum of the selected signals added in each group:
        # - signal added to group 1: group 1 signal 1 + group 2 signal 1
        # - signal added to group 2: group 1 signal 2 + group 2 signal 2
        assert len(model.get_groups()) == n_groups  # no new group
        titles = []
        for idx in range(1):
            pfx_orig = ", ".join(obj.short_id for obj in groups[idx][:2])
            titles.append((groups[idx][-1].title, f"Σ({pfx_orig})"))
        check_titles(f"Single operand mode Σ[{groups[1].title}]", titles)

    Conf.proc.operation_mode.set(original_mode)


def test_pairwise_operations_mode_compute_n1():
    """Run pairwise operations mode test scenario
    with compute_n1 operation (e.g. sum)"""
    original_mode = Conf.proc.operation_mode.get()
    Conf.proc.operation_mode.set("pairwise")

    with cdl_app_context(exec_loop=True):
        win = app.create(h5files=[get_test_fnames("reorder*")[0]], console=False)
        panel = win.signalpanel
        view, model = panel.objview, panel.objmodel

        # Checking the operation mode:
        assert is_pairwise_mode()

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select the two first groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        view.select_groups(groups)

        # Checking that each group contains the same number of signals (this is
        # required for pairwise operations - this part of the test is checking
        # if the data file is the one we expect)
        n_objects = len(groups[0])
        assert all(len(group) == n_objects for group in groups)

        # Perform a sum operation
        panel.processor.compute_sum()

        # Operation mode is now pairwise, so the sum operation is applied to the
        # selected groups, and we should have a new group with as many signals as
        # the original groups, each signal being the sum of the corresponding signals:
        # - signal 1: group 1 signal 1 + group 2 signal 1
        # - signal 2: group 1 signal 1 + group 2 signal 2
        # ...
        assert len(model.get_groups()) == n_groups + 1
        new_group = model.get_group_from_number(n_groups + 1)
        assert len(new_group.get_objects()) == n_objects
        titles = []
        for idx in range(len(groups[0])):
            obj = new_group[idx]
            pfx_orig = ", ".join(obj.short_id for obj in (grp[idx] for grp in groups))
            titles.append((obj.title, f"Σ({pfx_orig})"))
        check_titles(f"Pairwise operations mode Σ[{new_group.title}]", titles)

        # Remove new group
        view.select_groups([new_group])
        panel.remove_object(force=True)

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select two signals of the first two groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        objs = [groups[0][0]] + [groups[0][-1]] + groups[1][-2:]
        view.select_objects(objs)

        # Perform a sum operation
        panel.processor.compute_sum()

        # Operation mode is now pairwise, so the sum operation is applied to the
        # selected signals, and we should have a new group with as many signals as
        # the selected signals, each signal being the sum of the corresponding signals:
        # - signal 1: group 1 signal 1 + group 2 signal 1
        # - signal 2: group 1 signal 1 + group 2 signal 2
        # ...
        assert len(model.get_groups()) == n_groups + 1
        new_group = model.get_group_from_number(n_groups + 1)
        assert len(new_group) == 2  # 2 signals were selected
        titles = []
        for idx, obj in enumerate(new_group):
            pfx_orig = ", ".join(obj.short_id for obj in objs[idx::2])
            titles.append((obj.title, f"Σ({pfx_orig})"))
        check_titles(f"Pairwise operations mode Σ[{new_group.title}]", titles)

    Conf.proc.operation_mode.set(original_mode)


def test_single_operand_mode_compute_n1n():
    """Run single operand mode test scenario
    with compute_n1n operation (e.g. difference)"""
    original_mode = Conf.proc.operation_mode.get()
    Conf.proc.operation_mode.set("single")

    with cdl_app_context(exec_loop=True):
        win = app.create(h5files=[get_test_fnames("reorder*")[0]], console=False)
        panel = win.signalpanel
        view, model = panel.objview, panel.objmodel

        # Checking the operation mode:
        assert not is_pairwise_mode()

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select the two first groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        view.select_groups(groups)
        n_objects = [len(grp) for grp in groups]

        # Perform a difference operation with the first signal of the third group
        group3 = model.get_group_from_number(3)
        panel.processor.compute_difference(group3[0])

        # Default operation mode is single operand mode, so we should have new signals
        # in each selected group being the difference between the original signals and
        # the selected signal:
        # - in group 1:
        #   - signal 1: group 1 signal 1 - group 3 signal 1
        #   - signal 2: group 1 signal 2 - group 3 signal 1
        # - in group 2:
        #   - signal 1: group 2 signal 1 - group 3 signal 1
        #   - signal 2: group 2 signal 2 - group 3 signal 1
        assert len(model.get_groups()) == n_groups
        new_objs = []
        for i_group, group in enumerate(groups):
            titles = []
            for i_obj in range(n_objects[i_group]):
                obj = group[i_obj + n_objects[i_group]]
                titles.append(
                    (obj.title, f"{group[i_obj].short_id}-{group3[0].short_id}")
                )
                new_objs.append(obj)
            check_titles(f"Single operand mode Δ[{group.title}]", titles)

        # Remove new signals
        view.select_objects(new_objs)
        panel.remove_object(force=True)

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select the two first signals of the first two groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        objs = groups[0][:2] + groups[1][:2]
        view.select_objects(objs)
        n_objects = [2, 2]

        # Perform a difference operation with the first signal of the third group
        panel.processor.compute_difference(group3[0])

        # Default operation mode is single operand mode, so we should have new signals
        # being the difference between the original signals and the selected signal:
        # - in group 1:
        #   - signal 1: group 1 signal 1 - group 3 signal 1
        #   - signal 2: group 1 signal 2 - group 3 signal 1
        # - in group 2:
        #   - signal 1: group 2 signal 1 - group 3 signal 1
        #   - signal 2: group 2 signal 2 - group 3 signal 1
        assert len(model.get_groups()) == n_groups  # no new group
        for i_group, group in enumerate(groups):
            titles = []
            for i_obj in range(n_objects[i_group]):
                obj = group[len(group) - n_objects[i_group] + i_obj]
                titles.append(
                    (obj.title, f"{group[i_obj].short_id}-{group3[0].short_id}")
                )
            check_titles(f"Single operand mode Δ[{group.title}]", titles)

    Conf.proc.operation_mode.set(original_mode)


def test_pairwise_operations_mode_compute_n1n():
    """Run pairwise operations mode test scenario
    with compute_n1n operation (e.g. difference)"""
    original_mode = Conf.proc.operation_mode.get()
    Conf.proc.operation_mode.set("pairwise")

    with cdl_app_context(exec_loop=True):
        win = app.create(h5files=[get_test_fnames("reorder*")[0]], console=False)
        panel = win.signalpanel
        view, model = panel.objview, panel.objmodel

        # Checking the operation mode:
        assert is_pairwise_mode()

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select the two first groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        view.select_groups(groups)

        # Checking that each group contains the same number of signals (this is
        # required for pairwise operations - this part of the test is checking
        # if the data file is the one we expect)
        n_objects = len(groups[0])
        assert all(len(group) == n_objects for group in groups)

        # Perform a difference operation with the third group
        group3 = model.get_group_from_number(3)
        assert len(group3) == n_objects
        panel.processor.compute_difference(group3.get_objects())

        # Operation mode is now pairwise, so the difference operation is applied to the
        # selected groups, and we should have a new group with as many signals as the
        # original groups, each signal being the difference of the corresponding
        # signals:
        # - signal 1: group 1 signal 1 - group 2 signal 1
        # - signal 2: group 1 signal 1 - group 2 signal 2
        # ...
        assert len(model.get_groups()) == n_groups + 2
        new_groups = [
            model.get_group_from_number(idx) for idx in (n_groups + 1, n_groups + 2)
        ]
        execenv.print("Δ|pairwise")
        for i_new_grp, new_grp in enumerate(new_groups):
            assert len(new_grp.get_objects()) == n_objects
            titles = []
            for idx in range(n_objects):
                obj = new_grp[idx]
                obj1, obj2 = groups[i_new_grp][idx], group3[idx]
                titles.append((obj.title, f"{obj1.short_id}-{obj2.short_id}"))
            check_titles(f"Pairwise operations mode Δ[{new_grp.title}]", titles)

        # Remove new groups
        view.select_groups(new_groups)
        panel.remove_object(force=True)

        # Store the number of groups before the operations
        n_groups = len(model.get_groups())

        # Select two signals of the first two groups
        groups = [model.get_group_from_number(idx) for idx in (1, 2)]
        objs = [groups[0][0]] + [groups[0][-1]] + groups[1][-2:]
        view.select_objects(objs)
        n_objects = 2

        # Perform a difference operation with two signals from the third group
        objs2 = group3[:2]
        panel.processor.compute_difference(objs2)

        # Operation mode is now pairwise, so the difference operation is applied to the
        # selected signals, and we should have a new group with as many signals as the
        # selected signals, each signal being the difference of the corresponding
        # signals:
        # - signal 1: group 1 signal 1 - group 3 signal 1
        # - signal 2: group 1 signal 1 - group 3 signal 2
        # ...
        assert len(model.get_groups()) == n_groups + 2
        new_groups = [
            model.get_group_from_number(idx) for idx in (n_groups + 1, n_groups + 2)
        ]
        i_obj1 = 0
        execenv.print("Δ|pairwise")
        for i_new_grp, new_grp in enumerate(new_groups):
            assert len(new_grp.get_objects()) == n_objects
            titles = []
            for idx in range(n_objects):
                obj = new_grp[idx]
                obj1, obj2 = objs[i_obj1], objs2[idx]
                i_obj1 += 1
                titles.append((obj.title, f"{obj1.short_id}-{obj2.short_id}"))
            check_titles(f"Pairwise operations mode Δ[{new_grp.title}]", titles)

    Conf.proc.operation_mode.set(original_mode)


if __name__ == "__main__":
    test_single_operand_mode_compute_n1()
    test_pairwise_operations_mode_compute_n1()
    test_single_operand_mode_compute_n1n()
    test_pairwise_operations_mode_compute_n1n()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
I/O test

Testing DataLab specific formats.
"""

# guitest: show

from __future__ import annotations

import numpy as np
from guidata.qthelpers import qt_app_context

from cdl.core.io.image import funcs as image_funcs
from cdl.env import execenv
from cdl.obj import ImageObj, SignalObj, read_images, read_signals
from cdl.utils.tests import try_open_test_data
from cdl.utils.vistools import view_curve_items, view_images


def __read_objs(fname: str) -> list[ImageObj] | list[SignalObj]:
    """Read objects from a file"""
    if "curve_formats" in fname:
        objs = read_signals(fname)
    else:
        objs = read_images(fname)
    for obj in objs:
        if np.all(np.isnan(obj.data)):
            raise ValueError("Data is all NaNs")
    return objs


@try_open_test_data("Testing TXT file reader", "*.txt")
def open_txt(fname: str | None = None) -> None:
    """Testing TXT files"""
    objs = __read_objs(fname)
    for obj in objs:
        execenv.print(obj)
    view_curve_items([obj.make_item() for obj in objs])


@try_open_test_data("Testing CSV file reader", "*.csv")
def open_csv(fname: str | None = None) -> None:
    """Testing CSV files"""
    objs = __read_objs(fname)
    for obj in objs:
        execenv.print(obj)
    view_curve_items([obj.make_item() for obj in objs])


@try_open_test_data("Testing MAT-File reader", "*.mat")
def open_mat(fname: str | None = None) -> None:
    """Testing MAT files"""
    objs = __read_objs(fname)
    for obj in objs:
        execenv.print(obj)
    if isinstance(objs[0], SignalObj):
        view_curve_items([obj.make_item() for obj in objs])
    else:
        view_images([obj.data for obj in objs])


@try_open_test_data("Testing SIF file handler", "*.sif")
def open_sif(fname: str | None = None) -> None:
    """Testing SIF files"""
    execenv.print(image_funcs.SIFFile(fname))
    datalist = image_funcs.imread_sif(fname)
    view_images(datalist)


@try_open_test_data("Testing SCOR-DATA file handler", "*.scor-data")
def open_scordata(fname: str | None = None) -> None:
    """Testing SCOR-DATA files"""
    execenv.print(image_funcs.SCORFile(fname))
    data = image_funcs.imread_scor(fname)
    view_images(data)


def test_io1() -> None:
    """I/O test"""
    with qt_app_context():
        open_txt()
        open_csv()
        open_mat()
        open_sif()
        open_scordata()


if __name__ == "__main__":
    test_io1()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab launcher module
"""

from __future__ import annotations

from guidata.configtools import get_image_file_path
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from cdl.config import Conf
from cdl.core.gui.main import CDLMainWindow
from cdl.env import execenv
from cdl.utils.qthelpers import cdl_app_context


def create(
    splash: bool = True,
    console: bool | None = None,
    objects=None,
    h5files=None,
    size=None,
) -> CDLMainWindow:
    """Create DataLab application and return mainwindow instance

    Args:
        splash: if True, show splash screen
        console: if True, show console
        objects: list of objects to add to the mainwindow
        h5files: list of h5files to open
        size: mainwindow size (width, height)
    """
    if splash:
        # Showing splash screen
        pixmap = QG.QPixmap(get_image_file_path("DataLab-Splash.png"))
        splashscreen = QW.QSplashScreen(pixmap, QC.Qt.WindowStaysOnTopHint)
        splashscreen.show()
    window = CDLMainWindow(console=console)
    if size is not None:
        width, height = size
        window.resize(width, height)
    if splash:
        splashscreen.finish(window)
    if Conf.main.window_maximized.get(None):
        window.showMaximized()
    else:
        window.showNormal()
    if h5files is not None:
        window.open_h5_files(h5files, import_all=True)
    if objects is not None:
        for obj in objects:
            window.add_object(obj)
    if execenv.h5browser_file is not None:
        window.import_h5_file(execenv.h5browser_file)
    return window


def run(console=None, objects=None, h5files=None, size=None):
    """Run the DataLab application

    Note: this function is an entry point in `setup.py` and therefore
    may not be moved without modifying the package setup script.

    Args:
        console: if True, show console
        objects: list of objects to add to the mainwindow
        h5files: list of h5files to open
        size: mainwindow size (width, height)
    """
    if execenv.h5files:
        h5files = ([] if h5files is None else h5files) + execenv.h5files

    with cdl_app_context(exec_loop=True):
        window = create(
            splash=True, console=console, objects=objects, h5files=h5files, size=size
        )
        QW.QApplication.processEvents()
        window.execute_post_show_actions()


if __name__ == "__main__":
    run()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Unit tests for signal analysis features
---------------------------------------

Features from the "Analysis" menu are covered by this test.
The "Analysis" menu contains functions to compute signal properties like
bandwidth, ENOB, etc.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

from __future__ import annotations

import pytest

import cdl.computation.signal as cps
import cdl.obj
import cdl.param
from cdl.tests.data import get_test_signal
from cdl.utils.tests import check_scalar_result


@pytest.mark.validation
def test_signal_bandwidth_3db() -> None:
    """Validation test for the bandwidth computation."""
    obj = get_test_signal("bandwidth.txt")
    df = cps.compute_bandwidth_3db(obj).to_dataframe()
    check_scalar_result("Bandwitdh@-3dB", df.L[0], 39.0, rtol=0.001)


@pytest.mark.validation
def test_dynamic_parameters() -> None:
    """Validation test for dynamic parameters computation."""
    obj = get_test_signal("dynamic_parameters.txt")
    param = cdl.param.DynamicParam.create(full_scale=1.0)
    df = cps.compute_dynamic_parameters(obj, param).to_dataframe()
    check_scalar_result("ENOB", df.ENOB[0], 5.1, rtol=0.001)
    check_scalar_result("SINAD", df.SINAD[0], 32.49, rtol=0.001)
    check_scalar_result("THD", df.THD[0], -30.18, rtol=0.001)
    check_scalar_result("SFDR", df.SFDR[0], 34.03, rtol=0.001)
    check_scalar_result("Freq", df.Freq[0], 49998377.464, rtol=0.001)
    check_scalar_result("SNR", df.SNR[0], 101.52, rtol=0.001)


@pytest.mark.validation
def test_signal_sampling_rate_period() -> None:
    """Validation test for the sampling rate and period computation."""
    obj = get_test_signal("dynamic_parameters.txt")
    df = cps.compute_sampling_rate_period(obj).to_dataframe()
    check_scalar_result("Sampling rate", df["fs"][0], 1.0e10, rtol=0.001)
    check_scalar_result("Period", df["T"][0], 1.0e-10, rtol=0.001)


@pytest.mark.validation
def test_signal_contrast() -> None:
    """Validation test for the contrast computation."""
    obj = get_test_signal("fw1e2.txt")
    df = cps.compute_contrast(obj).to_dataframe()
    check_scalar_result("Contrast", df.contrast[0], 0.825, rtol=0.001)


@pytest.mark.validation
def test_signal_x_at_minmax() -> None:
    """Validation test for the x value at min/max computation."""
    obj = get_test_signal("fw1e2.txt")
    df = cps.compute_x_at_minmax(obj).to_dataframe()
    check_scalar_result("X@Ymin", df["X@Ymin"][0], 0.803, rtol=0.001)
    check_scalar_result("X@Ymax", df["X@Ymax"][0], 5.184, rtol=0.001)


if __name__ == "__main__":
    test_signal_bandwidth_3db()
    test_dynamic_parameters()
    test_signal_sampling_rate_period()
    test_signal_contrast()
    test_signal_x_at_minmax()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab base proxy module
-------------------------
"""

# How to add a new method to the proxy:
# -------------------------------------
#
# 1.  Add the method to the AbstractCDLControl class, as an abstract method
#
# 2a. If the method requires any data conversion to get through the XML-RPC layer,
#     implement the method in both LocalProxy and RemoteClient classes
#
# 2b. If the method does not require any data conversion, implement the method
#     directly in the BaseProxy class, so that it is available to both LocalProxy
#     and RemoteClient classes without any code duplication
#
# 3.  Implement the method in the CDLMainWindow class
#
# 4.  Implement the method in the RemoteServer class (it will be automatically
#     registered as an XML-RPC method, like all methods of AbstractCDLControl)

from __future__ import annotations

import abc
from collections.abc import Callable
from contextlib import contextmanager
from typing import TYPE_CHECKING, Generator

import guidata.dataset as gds
import numpy as np

from cdl.obj import ImageObj, SignalObj

if TYPE_CHECKING:
    from collections.abc import Iterator

    from cdl.core.gui.main import CDLMainWindow
    from cdl.core.remote import ServerProxy


class AbstractCDLControl(abc.ABC):
    """Abstract base class for controlling DataLab (main window or remote server)"""

    def __len__(self) -> int:
        """Return number of objects"""
        return len(self.get_object_uuids())

    def __getitem__(
        self,
        nb_id_title: int | str | None = None,
    ) -> SignalObj | ImageObj:
        """Return object"""
        return self.get_object(nb_id_title)

    def __iter__(self) -> Iterator[SignalObj | ImageObj]:
        """Iterate over objects"""
        uuids = self.get_object_uuids()
        for uuid in uuids:
            yield self.get_object(uuid)

    def __str__(self) -> str:
        """Return object string representation"""
        return super().__repr__()

    def __repr__(self) -> str:
        """Return object representation"""
        titles = self.get_object_titles()
        uuids = self.get_object_uuids()
        text = f"{str(self)} (DataLab, {len(titles)} items):\n"
        for uuid, title in zip(uuids, titles):
            text += f"  {uuid}: {title}\n"
        return text

    def __bool__(self) -> bool:
        """Return True if model is not empty"""
        return bool(self.get_object_uuids())

    def __contains__(self, id_title: str) -> bool:
        """Return True if object (UUID or title) is in model"""
        return id_title in (self.get_object_titles() + self.get_object_uuids())

    @classmethod
    def get_public_methods(cls) -> list[str]:
        """Return all public methods of the class, except itself.

        Returns:
            list[str]: List of public methods
        """
        return [
            method
            for method in dir(cls)
            if not method.startswith(("_", "context_"))
            and method != "get_public_methods"
        ]

    @abc.abstractmethod
    def get_version(self) -> str:
        """Return DataLab public version.

        Returns:
            str: DataLab version
        """

    @abc.abstractmethod
    def close_application(self) -> None:
        """Close DataLab application"""

    @abc.abstractmethod
    def raise_window(self) -> None:
        """Raise DataLab window"""

    @abc.abstractmethod
    def get_current_panel(self) -> str:
        """Return current panel name.

        Returns:
            str: Panel name (valid values: "signal", "image", "macro"))
        """

    @abc.abstractmethod
    def set_current_panel(self, panel: str) -> None:
        """Switch to panel.

        Args:
            panel (str): Panel name (valid values: "signal", "image", "macro"))
        """

    @abc.abstractmethod
    def reset_all(self) -> None:
        """Reset all application data"""

    @abc.abstractmethod
    def toggle_auto_refresh(self, state: bool) -> None:
        """Toggle auto refresh state.

        Args:
            state (bool): Auto refresh state
        """

    # Returns a context manager to temporarily disable autorefresh
    @contextmanager
    def context_no_refresh(self) -> Generator[None, None, None]:
        """Return a context manager to temporarily disable auto refresh.

        Returns:
            Context manager

        Example:

            >>> with proxy.context_no_refresh():
            ...     proxy.add_image("image1", data1)
            ...     proxy.compute_fft()
            ...     proxy.compute_wiener()
            ...     proxy.compute_ifft()
            ...     # Auto refresh is disabled during the above operations
        """
        self.toggle_auto_refresh(False)
        try:
            yield
        finally:
            self.toggle_auto_refresh(True)

    @abc.abstractmethod
    def toggle_show_titles(self, state: bool) -> None:
        """Toggle show titles state.

        Args:
            state (bool): Show titles state
        """

    @abc.abstractmethod
    def save_to_h5_file(self, filename: str) -> None:
        """Save to a DataLab HDF5 file.

        Args:
            filename (str): HDF5 file name
        """

    @abc.abstractmethod
    def open_h5_files(
        self,
        h5files: list[str] | None = None,
        import_all: bool | None = None,
        reset_all: bool | None = None,
    ) -> None:
        """Open a DataLab HDF5 file or import from any other HDF5 file.

        Args:
            h5files (list[str] | None): List of HDF5 files to open. Defaults to None.
            import_all (bool | None): Import all objects from HDF5 files.
                Defaults to None.
            reset_all (bool | None): Reset all application data. Defaults to None.
        """

    @abc.abstractmethod
    def import_h5_file(self, filename: str, reset_all: bool | None = None) -> None:
        """Open DataLab HDF5 browser to Import HDF5 file.

        Args:
            filename (str): HDF5 file name
            reset_all (bool | None): Reset all application data. Defaults to None.
        """

    @abc.abstractmethod
    def load_from_files(self, filenames: list[str]) -> None:
        """Open objects from files in current panel (signals/images).

        Args:
            filenames: list of file names
        """

    @abc.abstractmethod
    def add_signal(
        self,
        title: str,
        xdata: np.ndarray,
        ydata: np.ndarray,
        xunit: str | None = None,
        yunit: str | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
    ) -> bool:  # pylint: disable=too-many-arguments
        """Add signal data to DataLab.

        Args:
            title (str): Signal title
            xdata (numpy.ndarray): X data
            ydata (numpy.ndarray): Y data
            xunit (str | None): X unit. Defaults to None.
            yunit (str | None): Y unit. Defaults to None.
            xlabel (str | None): X label. Defaults to None.
            ylabel (str | None): Y label. Defaults to None.

        Returns:
            bool: True if signal was added successfully, False otherwise

        Raises:
            ValueError: Invalid xdata dtype
            ValueError: Invalid ydata dtype
        """

    @abc.abstractmethod
    def add_image(
        self,
        title: str,
        data: np.ndarray,
        xunit: str | None = None,
        yunit: str | None = None,
        zunit: str | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
        zlabel: str | None = None,
    ) -> bool:  # pylint: disable=too-many-arguments
        """Add image data to DataLab.

        Args:
            title (str): Image title
            data (numpy.ndarray): Image data
            xunit (str | None): X unit. Defaults to None.
            yunit (str | None): Y unit. Defaults to None.
            zunit (str | None): Z unit. Defaults to None.
            xlabel (str | None): X label. Defaults to None.
            ylabel (str | None): Y label. Defaults to None.
            zlabel (str | None): Z label. Defaults to None.

        Returns:
            bool: True if image was added successfully, False otherwise

        Raises:
            ValueError: Invalid data dtype
        """

    @abc.abstractmethod
    def add_object(self, obj: SignalObj | ImageObj) -> bool:
        """Add object to DataLab.

        Args:
            obj (SignalObj | ImageObj): Signal or image object

        Returns:
            bool: True if object was added successfully, False otherwise
        """

    @abc.abstractmethod
    def get_sel_object_uuids(self, include_groups: bool = False) -> list[str]:
        """Return selected objects uuids.

        Args:
            include_groups: If True, also return objects from selected groups.

        Returns:
            List of selected objects uuids.
        """

    @abc.abstractmethod
    def select_objects(
        self,
        selection: list[int | str],
        panel: str | None = None,
    ) -> None:
        """Select objects in current panel.

        Args:
            selection: List of object numbers (1 to N) or uuids to select
            panel: panel name (valid values: "signal", "image").
             If None, current panel is used. Defaults to None.
        """

    @abc.abstractmethod
    def select_groups(
        self, selection: list[int | str] | None = None, panel: str | None = None
    ) -> None:
        """Select groups in current panel.

        Args:
            selection: List of group numbers (1 to N), or list of group uuids,
             or None to select all groups. Defaults to None.
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used. Defaults to None.
        """

    @abc.abstractmethod
    def delete_metadata(
        self, refresh_plot: bool = True, keep_roi: bool = False
    ) -> None:
        """Delete metadata of selected objects

        Args:
            refresh_plot: Refresh plot. Defaults to True.
            keep_roi: Keep ROI. Defaults to False.
        """

    @abc.abstractmethod
    def get_group_titles_with_object_infos(
        self,
    ) -> tuple[list[str], list[list[str]], list[list[str]]]:
        """Return groups titles and lists of inner objects uuids and titles.

        Returns:
            Tuple: groups titles, lists of inner objects uuids and titles
        """

    @abc.abstractmethod
    def get_object_titles(self, panel: str | None = None) -> list[str]:
        """Get object (signal/image) list for current panel.
        Objects are sorted by group number and object index in group.

        Args:
            panel: panel name (valid values: "signal", "image", "macro").
             If None, current data panel is used (i.e. signal or image panel).

        Returns:
            List of object titles

        Raises:
            ValueError: if panel not found
        """

    @abc.abstractmethod
    def get_object(
        self,
        nb_id_title: int | str | None = None,
        panel: str | None = None,
    ) -> SignalObj | ImageObj:
        """Get object (signal/image) from index.

        Args:
            nb_id_title: Object number, or object id, or object title.
             Defaults to None (current object).
            panel: Panel name. Defaults to None (current panel).

        Returns:
            Object

        Raises:
            KeyError: if object not found
        """

    @abc.abstractmethod
    def get_object_uuids(self, panel: str | None = None) -> list[str]:
        """Get object (signal/image) uuid list for current panel.
        Objects are sorted by group number and object index in group.

        Args:
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used.

        Returns:
            list[str]: list of object uuids

        Raises:
            ValueError: if panel not found
        """

    @abc.abstractmethod
    def get_object_shapes(
        self,
        nb_id_title: int | str | None = None,
        panel: str | None = None,
    ) -> list:
        """Get plot item shapes associated to object (signal/image).

        Args:
            nb_id_title: Object number, or object id, or object title.
             Defaults to None (current object).
            panel: Panel name. Defaults to None (current panel).

        Returns:
            List of plot item shapes
        """

    @abc.abstractmethod
    def add_annotations_from_items(
        self, items: list, refresh_plot: bool = True, panel: str | None = None
    ) -> None:
        """Add object annotations (annotation plot items).

        Args:
            items (list): annotation plot items
            refresh_plot (bool | None): refresh plot. Defaults to True.
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used.
        """

    @abc.abstractmethod
    def add_label_with_title(
        self, title: str | None = None, panel: str | None = None
    ) -> None:
        """Add a label with object title on the associated plot

        Args:
            title (str | None): Label title. Defaults to None.
                If None, the title is the object title.
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used.
        """

    @abc.abstractmethod
    def run_macro(self, number_or_title: int | str | None = None) -> None:
        """Run macro.

        Args:
             number: Number of the macro (starting at 1). Defaults to None (run
              current macro, or does nothing if there is no macro).
        """

    @abc.abstractmethod
    def stop_macro(self, number_or_title: int | str | None = None) -> None:
        """Stop macro.

        Args:
            number: Number of the macro (starting at 1). Defaults to None (stop
             current macro, or does nothing if there is no macro).
        """

    @abc.abstractmethod
    def import_macro_from_file(self, filename: str) -> None:
        """Import macro from file

        Args:
            filename: Filename.
        """

    @abc.abstractmethod
    def calc(self, name: str, param: gds.DataSet | None = None) -> gds.DataSet:
        """Call compute function ``name`` in current panel's processor.

        Args:
            name: Compute function name
            param: Compute function parameter. Defaults to None.

        Raises:
            ValueError: unknown function
        """

    def __getattr__(self, name: str) -> Callable:
        """Return compute function ``name`` in current panel's processor.

        Args:
            name (str): Compute function name

        Returns:
            Callable: Compute function

        Raises:
            AttributeError: If compute function ``name`` does not exist
        """

        def compute_func(param: gds.DataSet | None = None) -> gds.DataSet:
            """Compute function.

            Args:
                param (guidata.dataset.DataSet | None): Compute function
                 parameter. Defaults to None.

            Returns:
                guidata.dataset.DataSet: Compute function result
            """
            return self.calc(name, param)

        if name.startswith("compute_"):
            return compute_func
        raise AttributeError(f"DataLab has no compute function '{name}'")


class BaseProxy(AbstractCDLControl, metaclass=abc.ABCMeta):
    """Common base class for DataLab proxies

    Args:
        cdl (CDLMainWindow | ServerProxy | None): CDLMainWindow instance or
            ServerProxy instance. If None, then the proxy implementation will
            have to set it later (e.g. see RemoteClient).
    """

    def __init__(self, cdl: CDLMainWindow | ServerProxy | None = None) -> None:
        self._cdl = cdl

    def get_version(self) -> str:
        """Return DataLab public version.

        Returns:
            str: DataLab version
        """
        return self._cdl.get_version()

    def close_application(self) -> None:
        """Close DataLab application"""
        self._cdl.close_application()

    def raise_window(self) -> None:
        """Raise DataLab window"""
        self._cdl.raise_window()

    def get_current_panel(self) -> str:
        """Return current panel name.

        Returns:
            str: Panel name (valid values: "signal", "image", "macro"))
        """
        return self._cdl.get_current_panel()

    def set_current_panel(self, panel: str) -> None:
        """Switch to panel.

        Args:
            panel (str): Panel name (valid values: "signal", "image", "macro"))
        """
        self._cdl.set_current_panel(panel)

    def reset_all(self) -> None:
        """Reset all application data"""
        self._cdl.reset_all()

    def toggle_auto_refresh(self, state: bool) -> None:
        """Toggle auto refresh state.

        Args:
            state (bool): Auto refresh state
        """
        self._cdl.toggle_auto_refresh(state)

    def toggle_show_titles(self, state: bool) -> None:
        """Toggle show titles state.

        Args:
            state (bool): Show titles state
        """
        self._cdl.toggle_show_titles(state)

    def save_to_h5_file(self, filename: str) -> None:
        """Save to a DataLab HDF5 file.

        Args:
            filename (str): HDF5 file name
        """
        self._cdl.save_to_h5_file(filename)

    def open_h5_files(
        self,
        h5files: list[str] | None = None,
        import_all: bool | None = None,
        reset_all: bool | None = None,
    ) -> None:
        """Open a DataLab HDF5 file or import from any other HDF5 file.

        Args:
            h5files (list[str] | None): List of HDF5 files to open. Defaults to None.
            import_all (bool | None): Import all objects from HDF5 files.
                Defaults to None.
            reset_all (bool | None): Reset all application data. Defaults to None.
        """
        self._cdl.open_h5_files(h5files, import_all, reset_all)

    def import_h5_file(self, filename: str, reset_all: bool | None = None) -> None:
        """Open DataLab HDF5 browser to Import HDF5 file.

        Args:
            filename (str): HDF5 file name
            reset_all (bool | None): Reset all application data. Defaults to None.
        """
        self._cdl.import_h5_file(filename, reset_all)

    def load_from_files(self, filenames: list[str]) -> None:
        """Open objects from files in current panel (signals/images).

        Args:
            filenames: list of file names
        """
        self._cdl.load_from_files(filenames)

    def get_sel_object_uuids(self, include_groups: bool = False) -> list[str]:
        """Return selected objects uuids.

        Args:
            include_groups: If True, also return objects from selected groups.

        Returns:
            List of selected objects uuids.
        """
        return self._cdl.get_sel_object_uuids(include_groups)

    def select_objects(
        self,
        selection: list[int | str],
        panel: str | None = None,
    ) -> None:
        """Select objects in current panel.

        Args:
            selection: List of object numbers (1 to N) or uuids to select
            panel: panel name (valid values: "signal", "image").
             If None, current panel is used. Defaults to None.
        """
        self._cdl.select_objects(selection, panel)

    def select_groups(
        self, selection: list[int | str] | None = None, panel: str | None = None
    ) -> None:
        """Select groups in current panel.

        Args:
            selection: List of group numbers (1 to N), or list of group uuids,
             or None to select all groups. Defaults to None.
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used. Defaults to None.
        """
        self._cdl.select_groups(selection, panel)

    def delete_metadata(
        self, refresh_plot: bool = True, keep_roi: bool = False
    ) -> None:
        """Delete metadata of selected objects

        Args:
            refresh_plot: Refresh plot. Defaults to True.
            keep_roi: Keep ROI. Defaults to False.
        """
        self._cdl.delete_metadata(refresh_plot, keep_roi)

    def get_group_titles_with_object_infos(
        self,
    ) -> tuple[list[str], list[list[str]], list[list[str]]]:
        """Return groups titles and lists of inner objects uuids and titles.

        Returns:
            Tuple: groups titles, lists of inner objects uuids and titles
        """
        return self._cdl.get_group_titles_with_object_infos()

    def get_object_titles(self, panel: str | None = None) -> list[str]:
        """Get object (signal/image) list for current panel.
        Objects are sorted by group number and object index in group.

        Args:
            panel: panel name (valid values: "signal", "image", "macro").
             If None, current data panel is used (i.e. signal or image panel).

        Returns:
            List of object titles

        Raises:
            ValueError: if panel not found
        """
        return self._cdl.get_object_titles(panel)

    def get_object_uuids(self, panel: str | None = None) -> list[str]:
        """Get object (signal/image) uuid list for current panel.
        Objects are sorted by group number and object index in group.

        Args:
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used.

        Returns:
            list[str]: list of object uuids

        Raises:
            ValueError: if panel not found
        """
        return self._cdl.get_object_uuids(panel)

    def add_label_with_title(
        self, title: str | None = None, panel: str | None = None
    ) -> None:
        """Add a label with object title on the associated plot

        Args:
            title (str | None): Label title. Defaults to None.
                If None, the title is the object title.
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used.
        """
        self._cdl.add_label_with_title(title, panel)

    def run_macro(self, number_or_title: int | str | None = None) -> None:
        """Run macro.

        Args:
            number_or_title: Macro number, or macro title.
             Defaults to None (current macro).

        Raises:
            ValueError: if macro not found
        """
        self._cdl.run_macro(number_or_title)

    def stop_macro(self, number_or_title: int | str | None = None) -> None:
        """Stop macro.

        Args:
            number_or_title: Macro number, or macro title.
             Defaults to None (current macro).

        Raises:
            ValueError: if macro not found
        """
        self._cdl.stop_macro(number_or_title)

    def import_macro_from_file(self, filename: str) -> None:
        """Import macro from file

        Args:
            filename: Filename.
        """
        return self._cdl.import_macro_from_file(filename)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Module for taking DataLab screenshots
"""

from cdl import config
from cdl.tests.features.applauncher import launcher1_app_test
from cdl.tests.scenarios import beautiful_app

if __name__ == "__main__":
    print("Updating screenshots...", end=" ")
    config.reset()  # Reset configuration (remove configuration file and initialize it)
    launcher1_app_test.test_launcher1(screenshots=True)
    config.reset()
    beautiful_app.run_beautiful_scenario(screenshots=True)
    print("done.")



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Native I/O module (native HDF5/JSON formats)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from guidata.io import HDF5Reader, HDF5Writer

import cdl

DATALAB_VERSION_NAME = "DataLab_Version"


class NativeH5Writer(HDF5Writer):
    """DataLab signal/image objects HDF5 guidata Dataset Writer class

    Args:
        filename (str): HDF5 file name
    """

    def __init__(self, filename: str) -> None:
        super().__init__(filename)
        self.h5[DATALAB_VERSION_NAME] = cdl.__version__


class NativeH5Reader(HDF5Reader):
    """DataLab signal/image objects HDF5 guidata dataset Writer class

    Args:
        filename (str): HDF5 file name
    """

    def __init__(self, filename: str) -> None:
        super().__init__(filename)
        self.version = self.h5[DATALAB_VERSION_NAME]



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Error message box test
"""

# guitest: show

from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtWidgets as QW

from cdl.widgets.warningerror import WarningErrorMessageBox


def error_message_box(category: str):
    """Test error message box

    Args:
        category (str): Error category
            Valid values are: "error", "warning"
    """
    with qt_app_context():
        win = QW.QMainWindow()
        win.setWindowTitle(f"DataLab {category.capitalize()} Message Box test")
        win.show()
        if category == "error":
            try:
                raise ValueError("Test error message box")
            except ValueError:
                context = "Test_error_message_box." * 5
                tip = "This error may occured when testing the error message box. " * 10
                dlg = WarningErrorMessageBox(win, "error", context, tip=tip)
                exec_dialog(dlg)
        elif category == "warning":
            context = "Test_warning_message_box." * 5
            message = "Test warning message box" * 10
            dlg = WarningErrorMessageBox(win, "warning", context, message)
            exec_dialog(dlg)
        else:
            raise ValueError(f"Invalid category: {category}")


def test_error_message_box():
    """Test error message box"""
    error_message_box("error")
    error_message_box("warning")


if __name__ == "__main__":
    test_error_message_box()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Image object and related classes
--------------------------------

"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

from __future__ import annotations

import abc
import enum
import re
from collections.abc import ByteString, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Generic, Literal, Type, Union
from uuid import uuid4

import guidata.dataset as gds
import numpy as np
from guidata.configtools import get_icon
from guidata.dataset import update_dataset
from plotpy.builder import make
from plotpy.items import (
    AnnotatedCircle,
    AnnotatedPolygon,
    AnnotatedRectangle,
    MaskedImageItem,
)
from skimage import draw

from cdl.algorithms.datatypes import clip_astype
from cdl.algorithms.image import scale_data_to_min_max
from cdl.config import Conf, _
from cdl.core.model import base

if TYPE_CHECKING:
    from qtpy import QtWidgets as QW


def to_builtin(obj) -> str | int | float | list | dict | np.ndarray | None:
    """Convert an object implementing a numeric value or collection
    into the corresponding builtin/NumPy type.

    Return None if conversion fails."""
    try:
        return int(obj) if int(obj) == float(obj) else float(obj)
    except (TypeError, ValueError):
        pass
    if isinstance(obj, ByteString):
        return str(obj)
    if isinstance(obj, Sequence):
        return str(obj) if len(obj) == len(str(obj)) else list(obj)
    if isinstance(obj, Mapping):
        return dict(obj)
    if isinstance(obj, np.ndarray):
        return obj
    return None


class ROI2DParam(base.BaseROIParam["ImageObj", "BaseSingleImageROI"]):
    """Image ROI parameters"""

    # Note: the ROI coordinates are expressed in pixel coordinates (integers)
    # => That is the only way to handle ROI parametrization for image objects.
    #    Otherwise, we would have to ask the user to systematically provide the
    #    physical coordinates: that would be cumbersome and error-prone.

    _geometry_prop = gds.GetAttrProp("geometry")
    _rfp = gds.FuncProp(_geometry_prop, lambda x: x != "rectangle")
    _cfp = gds.FuncProp(_geometry_prop, lambda x: x != "circle")
    _pfp = gds.FuncProp(_geometry_prop, lambda x: x != "polygon")

    # Do not declare it as a static method: not supported by Python 3.9
    def _lbl(name: str, index: int):  # pylint: disable=no-self-argument
        """Returns name<sub>index</sub>"""
        return f"{name}<sub>{index}</sub>"

    _ut = "pixels"

    geometries = ("rectangle", "circle", "polygon")
    geometry = gds.ChoiceItem(
        _("Geometry"), list(zip(geometries, geometries)), default="rectangle"
    ).set_prop("display", store=_geometry_prop, hide=True)

    # Parameters for rectangular ROI geometry:
    _tlcorner = gds.BeginGroup(_("Top left corner")).set_prop("display", hide=_rfp)
    x0 = gds.IntItem(_lbl("X", 0), unit=_ut).set_prop("display", hide=_rfp)
    y0 = gds.IntItem(_lbl("Y", 0), unit=_ut).set_pos(1).set_prop("display", hide=_rfp)
    _e_tlcorner = gds.EndGroup(_("Top left corner"))
    dx = gds.IntItem("ΔX", unit=_ut).set_prop("display", hide=_rfp)
    dy = gds.IntItem("ΔY", unit=_ut).set_pos(1).set_prop("display", hide=_rfp)

    # Parameters for circular ROI geometry:
    _cgroup = gds.BeginGroup(_("Center coordinates")).set_prop("display", hide=_cfp)
    xc = gds.IntItem(_lbl("X", "C"), unit=_ut).set_prop("display", hide=_cfp)
    yc = gds.IntItem(_lbl("Y", "C"), unit=_ut).set_pos(1).set_prop("display", hide=_cfp)
    _e_cgroup = gds.EndGroup(_("Center coordinates"))
    r = gds.IntItem(_("Radius"), unit=_ut).set_prop("display", hide=_cfp)

    # Parameters for polygonal ROI geometry:
    points = gds.FloatArrayItem(_("Coordinates") + f" ({_ut})").set_prop(
        "display", hide=_pfp
    )

    def to_single_roi(
        self, obj: ImageObj, title: str = ""
    ) -> PolygonalROI | RectangularROI | CircularROI:
        """Convert parameters to single ROI

        Args:
            obj: image object (used for conversion of pixel to physical coordinates)
            title: ROI title

        Returns:
            Single ROI
        """
        if self.geometry == "rectangle":
            return RectangularROI.from_param(obj, self)
        if self.geometry == "circle":
            return CircularROI.from_param(obj, self)
        if self.geometry == "polygon":
            return PolygonalROI.from_param(obj, self)
        raise ValueError(f"Unknown ROI geometry type: {self.geometry}")

    def get_suffix(self) -> str:
        """Get suffix text representation for ROI extraction"""
        if self.geometry == "rectangle":
            return f"x0={self.x0},y0={self.y0},dx={self.dx},dy={self.dy}"
        if self.geometry == "circle":
            return f"xc={self.xc},yc={self.yc},r={self.r}"
        if self.geometry == "polygon":
            return "polygon"
        raise ValueError(f"Unknown ROI geometry type: {self.geometry}")

    def get_extracted_roi(self, obj: ImageObj) -> ImageROI | None:
        """Get extracted ROI, i.e. the remaining ROI after extracting ROI from image.

        Args:
            obj: image object (used for conversion of pixel to physical coordinates)

        When extracting ROIs from an image to multiple images (i.e. one image per ROI),
        this method returns the ROI that has to be kept in the destination image. This
        is not necessary for a rectangular ROI: the destination image is simply a crop
        of the source image according to the ROI coordinates. But for a circular ROI or
        a polygonal ROI, the destination image is a crop of the source image according
        to the bounding box of the ROI. Thus, to avoid any loss of information, a ROI
        has to be defined for the destination image: this is the ROI returned by this
        method. It's simply the same as the source ROI, but with coordinates adjusted
        to the destination image. One may called this ROI the "extracted ROI".
        """
        if self.geometry == "rectangle":
            return None
        single_roi = self.to_single_roi(obj)
        x0, y0, _x1, _y1 = self.get_bounding_box_indices()
        single_roi.translate(obj, -x0, -y0)
        roi = ImageROI()
        roi.add_roi(single_roi)
        return roi

    def get_bounding_box_indices(self) -> tuple[int, int, int, int]:
        """Get bounding box (pixel coordinates)"""
        if self.geometry == "circle":
            x0, y0 = self.xc - self.r, self.yc - self.r
            x1, y1 = self.xc + self.r, self.yc + self.r
        elif self.geometry == "rectangle":
            x0, y0, x1, y1 = self.x0, self.y0, self.x0 + self.dx, self.y0 + self.dy
        else:
            self.points: np.ndarray
            x0, y0 = self.points[::2].min(), self.points[1::2].min()
            x1, y1 = self.points[::2].max(), self.points[1::2].max()
        return x0, y0, x1, y1

    def get_data(self, obj: ImageObj) -> np.ndarray:
        """Get data in ROI

        Args:
            obj: image object

        Returns:
            Data in ROI
        """
        x0, y0, x1, y1 = self.get_bounding_box_indices()
        x0, y0 = max(0, x0), max(0, y0)
        x1, y1 = min(obj.data.shape[1], x1), min(obj.data.shape[0], y1)
        return obj.data[y0:y1, x0:x1]


class BaseSingleImageROI(
    base.BaseSingleROI["ImageObj", ROI2DParam, base.TypeROIItem],
    Generic[base.TypeROIItem],
    abc.ABC,
):
    """Base class for single image ROI

    Args:
        coords: ROI edge coordinates (floats)
        title: ROI title

    .. note::

        The image ROI coords are expressed in physical coordinates (floats). The
        conversion to pixel coordinates is done in :class:`cdl.obj.ImageObj`
        (see :meth:`cdl.obj.ImageObj.physical_to_indices`). Most of the time,
        the physical coordinates are the same as the pixel coordinates, but this
        is not always the case (e.g. after image binning), so it's better to keep the
        physical coordinates in the ROI object: this will help reusing the ROI with
        different images (e.g. with different pixel sizes).
    """

    @abc.abstractmethod
    def get_bounding_box(self, obj: ImageObj) -> tuple[float, float, float, float]:
        """Get bounding box (physical coordinates)

        Args:
            obj: image object
        """

    @abc.abstractmethod
    def translate(self, obj: ImageObj, dx: int, dy: int) -> None:
        """Translate ROI

        Args:
            obj: image object
            dx: translation along X-axis
            dy: translation along Y-axis
        """


class PolygonalROI(BaseSingleImageROI[AnnotatedPolygon]):
    """Polygonal ROI

    Args:
        coords: ROI edge coordinates
        title: title

    Raises:
        ValueError: if number of coordinates is odd

    .. note:: The image ROI coords are expressed in physical coordinates (floats)
    """

    def check_coords(self) -> None:
        """Check if coords are valid

        Raises:
            ValueError: invalid coords
        """
        if len(self.coords) % 2 != 0:
            raise ValueError("Edge indices must be pairs of X, Y values")

    # pylint: disable=unused-argument
    @classmethod
    def from_param(cls: PolygonalROI, obj: ImageObj, param: ROI2DParam) -> PolygonalROI:
        """Create ROI from parameters

        Args:
            obj: image object
            param: parameters
        """
        indices = True  # ROI coordinates are in pixel coordinates in `ROI2DParam`
        return cls(param.points, indices=indices, title=param.get_title())

    def get_bounding_box(self, obj: ImageObj) -> tuple[float, float, float, float]:
        """Get bounding box (physical coordinates)

        Args:
            obj: image object
        """
        coords = self.get_physical_coords(obj)
        x_edges, y_edges = coords[::2], coords[1::2]
        return min(x_edges), min(y_edges), max(x_edges), max(y_edges)

    def translate(self, obj: ImageObj, dx: int, dy: int) -> None:
        """Translate ROI

        Args:
            obj: image object
            dx: translation along X-axis
            dy: translation along Y-axis
        """
        coords = self.get_indices_coords(obj)
        coords[::2] += int(dx)
        coords[1::2] += int(dy)
        self.set_indices_coords(obj, coords)

    def to_mask(self, obj: ImageObj) -> np.ndarray:
        """Create mask from ROI

        Args:
            obj: image object

        Returns:
            Mask (boolean array where True values are inside the ROI)
        """
        roi_mask = np.ones_like(obj.data, dtype=bool)
        indices = self.get_indices_coords(obj)
        rows, cols = indices[1::2], indices[::2]
        rr, cc = draw.polygon(rows, cols, shape=obj.data.shape)
        roi_mask[rr, cc] = False
        return roi_mask

    def to_param(self, obj: ImageObj, title: str | None = None) -> ROI2DParam:
        """Convert ROI to parameters

        Args:
            obj: object (image), for physical-indices coordinates conversion
            title: ROI title
        """
        param = ROI2DParam(title=self.title if title is None else title)
        param.geometry = "polygon"
        param.points = self.get_indices_coords(obj)
        return param

    def to_plot_item(self, obj: ImageObj, title: str | None = None) -> AnnotatedPolygon:
        """Make and return the annnotated polygon associated to ROI

        Args:
            obj: object (image), for physical-indices coordinates conversion
            title: title
        """
        item = AnnotatedPolygon(self.get_physical_coords(obj).reshape(-1, 2))
        item.annotationparam.title = self.title if title is None else title
        item.annotationparam.update_item(item)
        item.set_style("plot", "shape/drag")
        return item

    @classmethod
    def from_plot_item(cls: PolygonalROI, item: AnnotatedPolygon) -> PolygonalROI:
        """Create ROI from plot item

        Args:
            item: plot item
        """
        return cls(item.get_points().flatten(), False, item.annotationparam.title)


class RectangularROI(BaseSingleImageROI[AnnotatedRectangle]):
    """Rectangular ROI

    Args:
        coords: ROI edge coordinates (x0, y0, dx, dy)
        title: title

    .. note:: The image ROI coords are expressed in physical coordinates (floats)
    """

    def check_coords(self) -> None:
        """Check if coords are valid

        Raises:
            ValueError: invalid coords
        """
        if len(self.coords) != 4:
            raise ValueError("Rectangle ROI requires 4 coordinates")

    @classmethod
    def from_param(
        cls: RectangularROI, obj: ImageObj, param: ROI2DParam
    ) -> RectangularROI:
        """Create ROI from parameters

        Args:
            obj: image object
            param: parameters
        """
        ix0, iy0, ix1, iy1 = param.get_bounding_box_indices()
        coords = [ix0, iy0, ix1 - ix0, iy1 - iy0]
        indices = True  # ROI coordinates are in pixel coordinates in `ROI2DParam`
        return cls(coords, indices=indices, title=param.get_title())

    def get_bounding_box(self, obj: ImageObj) -> tuple[float, float, float, float]:
        """Get bounding box (physical coordinates)

        Args:
            obj: image object
        """
        x0, y0, dx, dy = self.get_physical_coords(obj)
        return x0, y0, x0 + dx, y0 + dy

    def translate(self, obj: ImageObj, dx: int, dy: int) -> None:
        """Translate ROI

        Args:
            obj: image object
            dx: translation along X-axis
            dy: translation along Y-axis
        """
        coords = self.get_indices_coords(obj)
        coords[0] += int(dx)
        coords[1] += int(dy)
        self.set_indices_coords(obj, coords)

    def get_physical_coords(self, obj: ImageObj) -> np.ndarray:
        """Return physical coords

        Args:
            obj: image object

        Returns:
            Physical coords
        """
        if self.indices:
            ix0, iy0, idx, idy = self.coords
            x0, y0, x1, y1 = obj.indices_to_physical([ix0, iy0, ix0 + idx, iy0 + idy])
            return [x0, y0, x1 - x0, y1 - y0]
        return self.coords

    def get_indices_coords(self, obj: ImageObj) -> np.ndarray:
        """Return indices coords

        Args:
            obj: image object

        Returns:
            Indices coords
        """
        if self.indices:
            return self.coords
        ix0, iy0, ix1, iy1 = obj.physical_to_indices(self.get_bounding_box(obj))
        return [ix0, iy0, ix1 - ix0, iy1 - iy0]

    def set_indices_coords(self, obj: ImageObj, coords: np.ndarray) -> None:
        """Set indices coords

        Args:
            obj: object (signal/image)
            coords: indices coords
        """
        if self.indices:
            self.coords = coords
        else:
            ix0, iy0, idx, idy = coords
            x0, y0, x1, y1 = obj.indices_to_physical([ix0, iy0, ix0 + idx, iy0 + idy])
            self.coords = [x0, y0, x1 - x0, y1 - y0]

    def to_mask(self, obj: ImageObj) -> np.ndarray:
        """Create mask from ROI

        Args:
            obj: image object

        Returns:
            Mask (boolean array where True values are inside the ROI)
        """
        roi_mask = np.ones_like(obj.data, dtype=bool)
        x0, y0, dx, dy = self.get_indices_coords(obj)
        roi_mask[max(y0, 0) : y0 + dy, max(x0, 0) : x0 + dx] = False
        return roi_mask

    def to_param(self, obj: ImageObj, title: str | None = None) -> ROI2DParam:
        """Convert ROI to parameters

        Args:
            obj: object (image), for physical-indices coordinates conversion
            title: ROI title
        """
        param = ROI2DParam(title=self.title if title is None else title)
        param.geometry = "rectangle"
        param.x0, param.y0, param.dx, param.dy = self.get_indices_coords(obj)
        return param

    def to_plot_item(
        self, obj: ImageObj, title: str | None = None
    ) -> AnnotatedRectangle:
        """Make and return the annnotated rectangle associated to ROI

        Args:
            obj: object (image), for physical-indices coordinates conversion
            title: title
        """

        def info_callback(item: AnnotatedRectangle) -> str:
            """Return info string for rectangular ROI"""
            x0, y0, x1, y1 = item.get_rect()
            if self.indices:
                x0, y0, x1, y1 = obj.physical_to_indices([x0, y0, x1, y1])
            x0, y0, dx, dy = self.rect_to_coords(x0, y0, x1, y1)
            return "<br>".join(
                [
                    f"X0, Y0 = {x0:g}, {y0:g}",
                    f"ΔX x ΔY  = {dx:g} x {dy:g}",
                ]
            )

        x0, y0, dx, dy = self.get_physical_coords(obj)
        x1, y1 = x0 + dx, y0 + dy
        title = self.title if title is None else title
        roi_item: AnnotatedRectangle = make.annotated_rectangle(x0, y0, x1, y1, title)
        roi_item.set_info_callback(info_callback)
        param = roi_item.label.labelparam
        param.anchor = "BL"
        param.xc, param.yc = 5, -5
        param.update_item(roi_item.label)
        return roi_item

    @staticmethod
    def rect_to_coords(
        x0: int | float, y0: int | float, x1: int | float, y1: int | float
    ) -> np.ndarray:
        """Convert rectangle to coordinates

        Args:
            x0: x0 (top-left corner)
            y0: y0 (top-left corner)
            x1: x1 (bottom-right corner)
            y1: y1 (bottom-right corner)

        Returns:
            Rectangle coordinates
        """
        return np.array([x0, y0, x1 - x0, y1 - y0], dtype=type(x0))

    @classmethod
    def from_plot_item(cls: RectangularROI, item: AnnotatedRectangle) -> RectangularROI:
        """Create ROI from plot item

        Args:
            item: plot item
        """
        rect = item.get_rect()
        return cls(cls.rect_to_coords(*rect), False, item.annotationparam.title)


class CircularROI(BaseSingleImageROI[AnnotatedCircle]):
    """Circular ROI

    Args:
        coords: ROI edge coordinates (xc, yc, r)
        title: title

    .. note:: The image ROI coords are expressed in physical coordinates (floats)
    """

    # pylint: disable=unused-argument
    @classmethod
    def from_param(cls: CircularROI, obj: ImageObj, param: ROI2DParam) -> CircularROI:
        """Create ROI from parameters

        Args:
            obj: image object
            param: parameters
        """
        ix0, iy0, ix1, iy1 = param.get_bounding_box_indices()
        ixc, iyc = (ix0 + ix1) * 0.5, (iy0 + iy1) * 0.5
        ir = (ix1 - ix0) * 0.5
        indices = True  # ROI coordinates are in pixel coordinates in `ROI2DParam`
        return cls([ixc, iyc, ir], indices=indices, title=param.get_title())

    def check_coords(self) -> None:
        """Check if coords are valid

        Raises:
            ValueError: invalid coords
        """
        if len(self.coords) != 3:
            raise ValueError("Circle ROI requires 3 coordinates")

    def get_bounding_box(self, obj: ImageObj) -> tuple[float, float, float, float]:
        """Get bounding box (physical coordinates)

        Args:
            obj: image object
        """
        xc, yc, r = self.get_physical_coords(obj)
        return xc - r, yc - r, xc + r, yc + r

    def translate(self, obj: ImageObj, dx: int, dy: int) -> None:
        """Translate ROI

        Args:
            obj: image object
            dx: translation along X-axis
            dy: translation along Y-axis
        """
        coords = self.get_indices_coords(obj)
        coords[0] += int(dx)
        coords[1] += int(dy)
        self.set_indices_coords(obj, coords)

    def get_physical_coords(self, obj: ImageObj) -> np.ndarray:
        """Return physical coords

        Args:
            obj: image object

        Returns:
            Physical coords
        """
        if self.indices:
            ixc, iyc, ir = self.coords
            x0, y0, x1, y1 = obj.indices_to_physical(
                [ixc - ir, iyc - ir, ixc + ir, iyc + ir]
            )
            return [0.5 * (x0 + x1), 0.5 * (y0 + y1), 0.5 * (x1 - x0)]
        return self.coords

    def get_indices_coords(self, obj: ImageObj) -> np.ndarray:
        """Return indices coords

        Args:
            obj: image object

        Returns:
            Indices coords
        """
        if self.indices:
            return self.coords
        ix0, iy0, ix1, iy1 = obj.physical_to_indices(self.get_bounding_box(obj))
        ixc, iyc = int((ix0 + ix1) * 0.5), int((iy0 + iy1) * 0.5)
        ir = int((ix1 - ix0) * 0.5)
        return [ixc, iyc, ir]

    def set_indices_coords(self, obj: ImageObj, coords: np.ndarray) -> None:
        """Set indices coords

        Args:
            obj: object (signal/image)
            coords: indices coords
        """
        if self.indices:
            self.coords = coords
        else:
            ixc, iyc, ir = coords
            x0, y0, x1, y1 = obj.indices_to_physical(
                [ixc - ir, iyc - ir, ixc + ir, iyc + ir]
            )
            self.coords = [0.5 * (x0 + x1), 0.5 * (y0 + y1), 0.5 * (x1 - x0)]

    def to_mask(self, obj: ImageObj) -> np.ndarray:
        """Create mask from ROI

        Args:
            obj: image object

        Returns:
            Mask (boolean array where True values are inside the ROI)
        """
        roi_mask = np.ones_like(obj.data, dtype=bool)
        ixc, iyc, ir = self.get_indices_coords(obj)
        yxratio = obj.dy / obj.dx
        rr, cc = draw.ellipse(iyc, ixc, ir / yxratio, ir, shape=obj.data.shape)
        roi_mask[rr, cc] = False
        return roi_mask

    def to_param(self, obj: ImageObj, title: str | None = None) -> ROI2DParam:
        """Convert ROI to parameters

        Args:
            obj: object (image), for physical-indices coordinates conversion
            title: ROI title
        """
        param = ROI2DParam(title=self.title if title is None else title)
        param.geometry = "circle"
        param.xc, param.yc, param.r = self.get_indices_coords(obj)
        return param

    def to_plot_item(self, obj: ImageObj, title: str | None = None) -> AnnotatedCircle:
        """Make and return the annnotated circle associated to ROI

        Args:
            obj: object (image), for physical-indices coordinates conversion
            title: title
        """

        def info_callback(item: AnnotatedCircle) -> str:
            """Return info string for circular ROI"""
            x0, y0, x1, y1 = item.get_rect()
            if self.indices:
                x0, y0, x1, y1 = obj.physical_to_indices([x0, y0, x1, y1])
            xc, yc, r = self.rect_to_coords(x0, y0, x1, y1)
            return "<br>".join(
                [
                    f"Center = {xc:g}, {yc:g}",
                    f"Radius = {r:g}",
                ]
            )

        xc, yc, r = self.get_physical_coords(obj)
        item = AnnotatedCircle(xc - r, yc, xc + r, yc)
        item.set_info_callback(info_callback)
        item.annotationparam.title = self.title if title is None else title
        item.annotationparam.update_item(item)
        item.set_style("plot", "shape/drag")
        return item

    @staticmethod
    def rect_to_coords(
        x0: int | float, y0: int | float, x1: int | float, y1: int | float
    ) -> np.ndarray:
        """Convert rectangle to circle coordinates

        Args:
            x0: x0 (top-left corner)
            y0: y0 (top-left corner)
            x1: x1 (bottom-right corner)
            y1: y1 (bottom-right corner)

        Returns:
            Circle coordinates
        """
        xc, yc, r = 0.5 * (x0 + x1), 0.5 * (y0 + y1), 0.5 * (x1 - x0)
        return np.array([xc, yc, r], dtype=type(x0))

    @classmethod
    def from_plot_item(cls: CircularROI, item: AnnotatedCircle) -> CircularROI:
        """Create ROI from plot item

        Args:
            item: plot item
        """
        rect = item.get_rect()
        return cls(cls.rect_to_coords(*rect), False, item.annotationparam.title)


class ImageROI(
    base.BaseROI[
        "ImageObj",
        BaseSingleImageROI,
        ROI2DParam,
        # `Union` is mandatory here for Python 3.9-3.10 compatibility:
        Union[AnnotatedPolygon, AnnotatedRectangle, AnnotatedCircle],
    ]
):
    """Image Regions of Interest

    Args:
        singleobj: if True, when extracting data defined by ROIs, only one object
         is created (default to True). If False, one object is created per single ROI.
         If None, the value is get from the user configuration
        inverse: if True, ROI is outside the region
    """

    PREFIX = "i"

    @staticmethod
    def get_compatible_single_roi_classes() -> list[Type[BaseSingleImageROI]]:
        """Return compatible single ROI classes"""
        return [RectangularROI, CircularROI, PolygonalROI]

    def to_mask(self, obj: ImageObj) -> np.ndarray[bool]:
        """Create mask from ROI

        Args:
            obj: image object

        Returns:
            Mask (boolean array where True values are inside the ROI)
        """
        mask = np.ones_like(obj.data, dtype=bool)
        for roi in self.single_rois:
            mask &= roi.to_mask(obj)
        return mask


def create_image_roi(
    geometry: Literal["rectangle", "circle", "polygon"],
    coords: np.ndarray | list[float] | list[list[float]],
    indices: bool = True,
    singleobj: bool | None = None,
    inverse: bool = False,
    title: str = "",
) -> ImageROI:
    """Create Image Regions of Interest (ROI) object.
    More ROIs can be added to the object after creation, using the `add_roi` method.

    Args:
        geometry: ROI type ('rectangle', 'circle', 'polygon')
        coords: ROI coords (physical coordinates), `[x0, y0, dx, dy]` for a rectangle,
         `[xc, yc, r]` for a circle, or `[x0, y0, x1, y1, ...]` for a polygon (lists or
         NumPy arrays are accepted). For multiple ROIs, nested lists or NumPy arrays are
         accepted but with a common geometry type (e.g.
         `[[xc1, yc1, r1], [xc2, yc2, r2], ...]` for circles).
        indices: if True, coordinates are indices, if False, they are physical values
         (default to True for images)
        singleobj: if True, when extracting data defined by ROIs, only one object
         is created (default to True). If False, one object is created per single ROI.
         If None, the value is get from the user configuration
        inverse: if True, ROI is outside the region
        title: title

    Returns:
        Regions of Interest (ROI) object

    Raises:
        ValueError: if ROI type is unknown or if the number of coordinates is invalid
    """
    coords = np.array(coords, float)
    if coords.ndim == 1:
        coords = coords.reshape(1, -1)
    roi = ImageROI(singleobj, inverse)
    if geometry == "rectangle":
        if coords.shape[1] != 4:
            raise ValueError("Rectangle ROI requires 4 coordinates")
        for row in coords:
            roi.add_roi(RectangularROI(row, indices, title))
    elif geometry == "circle":
        if coords.shape[1] != 3:
            raise ValueError("Circle ROI requires 3 coordinates")
        for row in coords:
            roi.add_roi(CircularROI(row, indices, title))
    elif geometry == "polygon":
        if coords.shape[1] % 2 != 0:
            raise ValueError("Polygon ROI requires pairs of X, Y coordinates")
        for row in coords:
            roi.add_roi(PolygonalROI(row, indices, title))
    else:
        raise ValueError(f"Unknown ROI type: {geometry}")
    return roi


class ImageObj(gds.DataSet, base.BaseObj[ImageROI, MaskedImageItem]):
    """Image object"""

    PREFIX = "i"
    CONF_FMT = Conf.view.ima_format
    DEFAULT_FMT = ".1f"
    VALID_DTYPES = (
        np.uint8,
        np.uint16,
        np.int16,
        np.int32,
        np.float32,
        np.float64,
        np.complex128,
    )

    def __init__(self, title=None, comment=None, icon=""):
        """Constructor

        Args:
            title: title
            comment: comment
            icon: icon
        """
        gds.DataSet.__init__(self, title, comment, icon)
        base.BaseObj.__init__(self)
        self.regenerate_uuid()
        self._dicom_template = None

    @staticmethod
    def get_roi_class() -> Type[ImageROI]:
        """Return ROI class"""
        return ImageROI

    def regenerate_uuid(self):
        """Regenerate UUID

        This method is used to regenerate UUID after loading the object from a file.
        This is required to avoid UUID conflicts when loading objects from file
        without clearing the workspace first.
        """
        self.uuid = str(uuid4())

    def __add_metadata(self, key: str, value: Any) -> None:
        """Add value to metadata if value can be converted into builtin/NumPy type

        Args:
            key: key
            value: value
        """
        stored_val = to_builtin(value)
        if stored_val is not None:
            self.metadata[key] = stored_val

    def set_metadata_from(self, obj: Mapping | dict) -> None:
        """Set metadata from object: dict-like (only string keys are considered)
        or any other object (iterating over supported attributes)

        Args:
            obj: object
        """
        self.reset_metadata_to_defaults()
        ptn = r"__[\S_]*__$"
        if isinstance(obj, Mapping):
            for key, value in obj.items():
                if isinstance(key, str) and not re.match(ptn, key):
                    self.__add_metadata(key, value)
        else:
            for attrname in dir(obj):
                if attrname != "GroupLength" and not re.match(ptn, attrname):
                    try:
                        attr = getattr(obj, attrname)
                        if not callable(attr) and attr:
                            self.__add_metadata(attrname, attr)
                    except AttributeError:
                        pass

    @property
    def dicom_template(self):
        """Get DICOM template"""
        return self._dicom_template

    @dicom_template.setter
    def dicom_template(self, template):
        """Set DICOM template"""
        if template is not None:
            ipp = getattr(template, "ImagePositionPatient", None)
            if ipp is not None:
                self.x0, self.y0 = float(ipp[0]), float(ipp[1])
            pxs = getattr(template, "PixelSpacing", None)
            if pxs is not None:
                self.dy, self.dx = float(pxs[0]), float(pxs[1])
            self.set_metadata_from(template)
            self._dicom_template = template

    uuid = gds.StringItem("UUID").set_prop("display", hide=True)

    _tabs = gds.BeginTabGroup("all")

    _datag = gds.BeginGroup(_("Data"))
    data = gds.FloatArrayItem(_("Data"))
    metadata = gds.DictItem(_("Metadata"), default={})
    _e_datag = gds.EndGroup(_("Data"))

    _dxdyg = gds.BeginGroup(f'{_("Origin")} / {_("Pixel spacing")}')
    _origin = gds.BeginGroup(_("Origin"))
    x0 = gds.FloatItem("X<sub>0</sub>", default=0.0)
    y0 = gds.FloatItem("Y<sub>0</sub>", default=0.0).set_pos(col=1)
    _e_origin = gds.EndGroup(_("Origin"))
    _pixel_spacing = gds.BeginGroup(_("Pixel spacing"))
    dx = gds.FloatItem("Δx", default=1.0, nonzero=True)
    dy = gds.FloatItem("Δy", default=1.0, nonzero=True).set_pos(col=1)
    _e_pixel_spacing = gds.EndGroup(_("Pixel spacing"))
    _e_dxdyg = gds.EndGroup(f'{_("Origin")} / {_("Pixel spacing")}')

    _unitsg = gds.BeginGroup(f'{_("Titles")} / {_("Units")}')
    title = gds.StringItem(_("Image title"), default=_("Untitled"))
    _tabs_u = gds.BeginTabGroup("units")
    _unitsx = gds.BeginGroup(_("X-axis"))
    xlabel = gds.StringItem(_("Title"), default="")
    xunit = gds.StringItem(_("Unit"), default="")
    _e_unitsx = gds.EndGroup(_("X-axis"))
    _unitsy = gds.BeginGroup(_("Y-axis"))
    ylabel = gds.StringItem(_("Title"), default="")
    yunit = gds.StringItem(_("Unit"), default="")
    _e_unitsy = gds.EndGroup(_("Y-axis"))
    _unitsz = gds.BeginGroup(_("Z-axis"))
    zlabel = gds.StringItem(_("Title"), default="")
    zunit = gds.StringItem(_("Unit"), default="")
    _e_unitsz = gds.EndGroup(_("Z-axis"))
    _e_tabs_u = gds.EndTabGroup("units")
    _e_unitsg = gds.EndGroup(f'{_("Titles")} / {_("Units")}')

    _scalesg = gds.BeginGroup(_("Scales"))
    _prop_autoscale = gds.GetAttrProp("autoscale")
    autoscale = gds.BoolItem(_("Auto scale"), default=True).set_prop(
        "display", store=_prop_autoscale
    )
    _tabs_b = gds.BeginTabGroup("bounds")
    _boundsx = gds.BeginGroup(_("X-axis"))
    xscalelog = gds.BoolItem(_("Logarithmic scale"), default=False)
    xscalemin = gds.FloatItem(_("Lower bound"), check=False).set_prop(
        "display", active=gds.NotProp(_prop_autoscale)
    )
    xscalemax = gds.FloatItem(_("Upper bound"), check=False).set_prop(
        "display", active=gds.NotProp(_prop_autoscale)
    )
    _e_boundsx = gds.EndGroup(_("X-axis"))
    _boundsy = gds.BeginGroup(_("Y-axis"))
    yscalelog = gds.BoolItem(_("Logarithmic scale"), default=False)
    yscalemin = gds.FloatItem(_("Lower bound"), check=False).set_prop(
        "display", active=gds.NotProp(_prop_autoscale)
    )
    yscalemax = gds.FloatItem(_("Upper bound"), check=False).set_prop(
        "display", active=gds.NotProp(_prop_autoscale)
    )
    _e_boundsy = gds.EndGroup(_("Y-axis"))
    _boundsz = gds.BeginGroup(_("LUT range"))
    zscalemin = gds.FloatItem(_("Lower bound"), check=False)
    zscalemax = gds.FloatItem(_("Upper bound"), check=False)
    _e_boundsz = gds.EndGroup(_("LUT range"))
    _e_tabs_b = gds.EndTabGroup("bounds")
    _e_scalesg = gds.EndGroup(_("Scales"))

    _e_tabs = gds.EndTabGroup("all")

    @property
    def width(self) -> float:
        """Return image width, i.e. number of columns multiplied by pixel size"""
        return self.data.shape[1] * self.dx

    @property
    def height(self) -> float:
        """Return image height, i.e. number of rows multiplied by pixel size"""
        return self.data.shape[0] * self.dy

    @property
    def xc(self) -> float:
        """Return image center X-axis coordinate"""
        return self.x0 + 0.5 * self.width

    @property
    def yc(self) -> float:
        """Return image center Y-axis coordinate"""
        return self.y0 + 0.5 * self.height

    def get_data(self, roi_index: int | None = None) -> np.ndarray:
        """
        Return original data (if ROI is not defined or `roi_index` is None),
        or ROI data (if both ROI and `roi_index` are defined).

        Args:
            roi_index: ROI index

        Returns:
            Masked data
        """
        if self.roi is None or roi_index is None:
            return self.data
        single_roi = self.roi.get_single_roi(roi_index)
        x0, y0, x1, y1 = self.physical_to_indices(single_roi.get_bounding_box(self))
        return self.get_masked_view()[y0:y1, x0:x1]

    def copy(self, title: str | None = None, dtype: np.dtype | None = None) -> ImageObj:
        """Copy object.

        Args:
            title: title
            dtype: data type

        Returns:
            Copied object
        """
        title = self.title if title is None else title
        obj = ImageObj(title=title)
        obj.title = title
        obj.xlabel = self.xlabel
        obj.ylabel = self.ylabel
        obj.xunit = self.xunit
        obj.yunit = self.yunit
        obj.zunit = self.zunit
        obj.x0 = self.x0
        obj.y0 = self.y0
        obj.dx = self.dx
        obj.dy = self.dy
        obj.metadata = base.deepcopy_metadata(self.metadata)
        obj.data = np.array(self.data, copy=True, dtype=dtype)
        obj.dicom_template = self.dicom_template
        return obj

    def set_data_type(self, dtype: np.dtype) -> None:
        """Change data type.
        If data type is integer, clip values to the new data type's range, thus avoiding
        overflow or underflow.

        Args:
            Data type
        """
        self.data = clip_astype(self.data, dtype)

    def __viewable_data(self) -> np.ndarray:
        """Return viewable data"""
        data = self.data.real
        if np.any(np.isnan(data)):
            data = np.nan_to_num(data, posinf=0, neginf=0)
        return data

    def update_plot_item_parameters(self, item: MaskedImageItem) -> None:
        """Update plot item parameters from object data/metadata

        Takes into account a subset of plot item parameters. Those parameters may
        have been overriden by object metadata entries or other object data. The goal
        is to update the plot item accordingly.

        This is *almost* the inverse operation of `update_metadata_from_plot_item`.

        Args:
            item: plot item
        """
        for axis in ("x", "y", "z"):
            unit = getattr(self, axis + "unit")
            fmt = r"%.1f"
            if unit:
                fmt = r"%.1f (" + unit + ")"
            setattr(item.param, axis + "format", fmt)
        # Updating origin and pixel spacing
        has_origin = self.x0 is not None and self.y0 is not None
        has_pixelspacing = self.dx is not None and self.dy is not None
        if has_origin or has_pixelspacing:
            x0, y0, dx, dy = 0.0, 0.0, 1.0, 1.0
            if has_origin:
                x0, y0 = self.x0, self.y0
            if has_pixelspacing:
                dx, dy = self.dx, self.dy
            shape = self.data.shape
            item.param.xmin, item.param.xmax = x0, x0 + dx * shape[1]
            item.param.ymin, item.param.ymax = y0, y0 + dy * shape[0]
        zmin, zmax = item.get_lut_range()
        if self.zscalemin is not None or self.zscalemax is not None:
            zmin = zmin if self.zscalemin is None else self.zscalemin
            zmax = zmax if self.zscalemax is None else self.zscalemax
            item.set_lut_range([zmin, zmax])
        super().update_plot_item_parameters(item)

    def update_metadata_from_plot_item(self, item: MaskedImageItem) -> None:
        """Update metadata from plot item.

        Takes into account a subset of plot item parameters. Those parameters may
        have been modified by the user through the plot item GUI. The goal is to
        update the metadata accordingly.

        This is *almost* the inverse operation of `update_plot_item_parameters`.

        Args:
            item: plot item
        """
        super().update_metadata_from_plot_item(item)
        # Updating the LUT range:
        self.zscalemin, self.zscalemax = item.get_lut_range()
        # Updating origin and pixel spacing:
        shape = self.data.shape
        param = item.param
        xmin, xmax, ymin, ymax = param.xmin, param.xmax, param.ymin, param.ymax
        if xmin == 0 and ymin == 0 and xmax == shape[1] and ymax == shape[0]:
            self.x0, self.y0, self.dx, self.dy = 0.0, 0.0, 1.0, 1.0
        else:
            self.x0, self.y0 = xmin, ymin
            self.dx, self.dy = (xmax - xmin) / shape[1], (ymax - ymin) / shape[0]

    def make_item(self, update_from: MaskedImageItem | None = None) -> MaskedImageItem:
        """Make plot item from data.

        Args:
            update_from: update from plot item

        Returns:
            Plot item
        """
        data = self.__viewable_data()
        item = make.maskedimage(
            data,
            self.maskdata,
            title=self.title,
            colormap="viridis",
            eliminate_outliers=Conf.view.ima_eliminate_outliers.get(),
            interpolation="nearest",
            show_mask=True,
        )
        if update_from is None:
            self.update_plot_item_parameters(item)
        else:
            update_dataset(item.param, update_from.param)
            item.param.update_item(item)
        return item

    def update_item(self, item: MaskedImageItem, data_changed: bool = True) -> None:
        """Update plot item from data.

        Args:
            item: plot item
            data_changed: if True, data has changed
        """
        if data_changed:
            item.set_data(self.__viewable_data(), lut_range=[item.min, item.max])
        item.set_mask(self.maskdata)
        item.param.label = self.title
        self.update_plot_item_parameters(item)
        item.plot().update_colormap_axis(item)

    def physical_to_indices(self, coords: list[float]) -> np.ndarray:
        """Convert coordinates from physical (real world) to (array) indices (pixel)

        Args:
            coords: coordinates

        Returns:
            Indices
        """
        indices = np.array(coords, float)
        ndim = indices.ndim
        if ndim == 1:
            indices = indices.reshape(1, -1)
        if indices.size > 0:
            indices[:, ::2] -= self.x0 + 0.5 * self.dx
            indices[:, ::2] /= self.dx
            indices[:, 1::2] -= self.y0 + 0.5 * self.dy
            indices[:, 1::2] /= self.dy
        if ndim == 1:
            indices = indices.flatten()
        return np.array(indices, int)

    def indices_to_physical(
        self, indices: list[float | int] | np.ndarray
    ) -> np.ndarray:
        """Convert coordinates from (array) indices to physical (real world)

        Args:
            indices: indices

        Returns:
            Coordinates
        """
        coords = np.array(indices, float)
        ndim = coords.ndim
        if ndim == 1:
            coords = coords.reshape(1, -1)
        if coords.size > 0:
            coords[:, ::2] *= self.dx
            coords[:, ::2] += self.x0 + 0.5 * self.dx
            coords[:, 1::2] *= self.dy
            coords[:, 1::2] += self.y0 + 0.5 * self.dy
        if ndim == 1:
            coords = coords.flatten()
        return coords

    def add_label_with_title(self, title: str | None = None) -> None:
        """Add label with title annotation

        Args:
            title: title (if None, use image title)
        """
        title = self.title if title is None else title
        if title:
            label = make.label(title, (self.x0, self.y0), (10, 10), "TL")
            self.add_annotations_from_items([label])


def create_image(
    title: str,
    data: np.ndarray | None = None,
    metadata: dict | None = None,
    units: tuple | None = None,
    labels: tuple | None = None,
) -> ImageObj:
    """Create a new Image object

    Args:
        title: image title
        data: image data
        metadata: image metadata
        units: X, Y, Z units (tuple of strings)
        labels: X, Y, Z labels (tuple of strings)

    Returns:
        Image object
    """
    assert isinstance(title, str)
    assert data is None or isinstance(data, np.ndarray)
    image = ImageObj(title=title)
    image.title = title
    image.data = data
    if units is not None:
        image.xunit, image.yunit, image.zunit = units
    if labels is not None:
        image.xlabel, image.ylabel, image.zlabel = labels
    if metadata is not None:
        image.metadata.update(metadata)
    return image


class ImageDatatypes(base.Choices):
    """Image data types"""

    @classmethod
    def from_dtype(cls, dtype):
        """Return member from NumPy dtype"""
        return getattr(cls, str(dtype).upper(), cls.UINT8)

    @classmethod
    def check(cls):
        """Check if data types are valid"""
        for member in cls:
            assert hasattr(np, member.value)

    #: Unsigned integer number stored with 8 bits
    UINT8 = enum.auto()
    #: Unsigned integer number stored with 16 bits
    UINT16 = enum.auto()
    #: Signed integer number stored with 16 bits
    INT16 = enum.auto()
    #: Float number stored with 32 bits
    FLOAT32 = enum.auto()
    #: Float number stored with 64 bits
    FLOAT64 = enum.auto()


ImageDatatypes.check()


class ImageTypes(base.Choices):
    """Image types"""

    #: Image filled with zeros
    ZEROS = _("zeros")
    #: Empty image (filled with data from memory state)
    EMPTY = _("empty")
    #: 2D Gaussian image
    GAUSS = _("gaussian")
    #: Image filled with random data (uniform law)
    UNIFORMRANDOM = _("random (uniform law)")
    #: Image filled with random data (normal law)
    NORMALRANDOM = _("random (normal law)")


class NewImageParam(gds.DataSet):
    """New image dataset"""

    hide_image_dtype = False
    hide_image_type = False

    title = gds.StringItem(_("Title"))
    height = gds.IntItem(
        _("Height"), help=_("Image height (total number of rows)"), min=1
    )
    width = gds.IntItem(
        _("Width"), help=_("Image width (total number of columns)"), min=1
    )
    dtype = gds.ChoiceItem(_("Data type"), ImageDatatypes.get_choices()).set_prop(
        "display", hide=gds.GetAttrProp("hide_image_dtype")
    )
    itype = gds.ChoiceItem(_("Type"), ImageTypes.get_choices()).set_prop(
        "display", hide=gds.GetAttrProp("hide_image_type")
    )


DEFAULT_TITLE = _("Untitled image")


def new_image_param(
    title: str | None = None,
    itype: ImageTypes | None = None,
    height: int | None = None,
    width: int | None = None,
    dtype: ImageDatatypes | None = None,
) -> NewImageParam:
    """Create a new Image dataset instance.

    Args:
        title: dataset title (default: None, uses default title)
        itype: image type (default: None, uses default type)
        height: image height (default: None, uses default height)
        width: image width (default: None, uses default width)
        dtype: image data type (default: None, uses default data type)

    Returns:
        New image dataset instance
    """
    title = DEFAULT_TITLE if title is None else title
    param = NewImageParam(title=title, icon=get_icon("new_image.svg"))
    param.title = title
    if height is not None:
        param.height = height
    if width is not None:
        param.width = width
    if dtype is not None:
        param.dtype = dtype
    if itype is not None:
        param.itype = itype
    return param


IMG_NB = 0


class Gauss2DParam(gds.DataSet):
    """2D Gaussian parameters"""

    a = gds.FloatItem("Norm")
    xmin = gds.FloatItem("Xmin", default=-10).set_pos(col=1)
    sigma = gds.FloatItem("σ", default=1.0)
    xmax = gds.FloatItem("Xmax", default=10).set_pos(col=1)
    mu = gds.FloatItem("μ", default=0.0)
    ymin = gds.FloatItem("Ymin", default=-10).set_pos(col=1)
    x0 = gds.FloatItem("X0", default=0)
    ymax = gds.FloatItem("Ymax", default=10).set_pos(col=1)
    y0 = gds.FloatItem("Y0", default=0).set_pos(col=0, colspan=1)


def create_image_from_param(
    newparam: NewImageParam,
    addparam: gds.DataSet | None = None,
    edit: bool = False,
    parent: QW.QWidget | None = None,
) -> ImageObj | None:
    """Create a new Image object from dialog box.

    Args:
        newparam: new image parameters
        addparam: additional parameters
        edit: Open a dialog box to edit parameters (default: False)
        parent: parent widget

    Returns:
        New image object or None if user cancelled
    """
    global IMG_NB  # pylint: disable=global-statement
    if newparam is None:
        newparam = new_image_param()
    if newparam.height is None:
        newparam.height = 500
    if newparam.width is None:
        newparam.width = 500
    if newparam.dtype is None:
        newparam.dtype = ImageDatatypes.UINT16
    incr_sig_nb = not newparam.title
    if incr_sig_nb:
        newparam.title = f"{newparam.title} {IMG_NB + 1:d}"
    if not edit or addparam is not None or newparam.edit(parent=parent):
        prefix = newparam.itype.name.lower()
        if incr_sig_nb:
            IMG_NB += 1
        image = create_image(newparam.title)
        shape = (newparam.height, newparam.width)
        dtype = newparam.dtype.value
        p = addparam
        if newparam.itype == ImageTypes.ZEROS:
            image.data = np.zeros(shape, dtype=dtype)
        elif newparam.itype == ImageTypes.EMPTY:
            image.data = np.empty(shape, dtype=dtype)
        elif newparam.itype == ImageTypes.GAUSS:
            if p is None:
                p = Gauss2DParam(_("2D-gaussian image"))
            if p.a is None:
                try:
                    p.a = np.iinfo(dtype).max / 2.0
                except ValueError:
                    p.a = 10.0
            if edit and not p.edit(parent=parent):
                return None
            x, y = np.meshgrid(
                np.linspace(p.xmin, p.xmax, shape[1]),
                np.linspace(p.ymin, p.ymax, shape[0]),
            )
            zgauss = p.a * np.exp(
                -((np.sqrt((x - p.x0) ** 2 + (y - p.y0) ** 2) - p.mu) ** 2)
                / (2.0 * p.sigma**2)
            )
            image.data = np.array(zgauss, dtype=dtype)
            if image.title == DEFAULT_TITLE:
                image.title = (
                    f"{prefix}(a={p.a:g},μ={p.mu:g},σ={p.sigma:g}),"
                    f"x0={p.x0:g},y0={p.y0:g})"
                )
        elif newparam.itype in (ImageTypes.UNIFORMRANDOM, ImageTypes.NORMALRANDOM):
            pclass = {
                ImageTypes.UNIFORMRANDOM: base.UniformRandomParam,
                ImageTypes.NORMALRANDOM: base.NormalRandomParam,
            }[newparam.itype]
            if p is None:
                p = pclass(_("Image") + " - " + newparam.itype.value)
                p.set_from_datatype(dtype)
            if edit and not p.edit(parent=parent):
                return None
            rng = np.random.default_rng(p.seed)
            if newparam.itype == ImageTypes.UNIFORMRANDOM:
                data = rng.random(shape)
                image.data = scale_data_to_min_max(data, p.vmin, p.vmax)
                if image.title == DEFAULT_TITLE:
                    image.title = (
                        f"{prefix}(vmin={p.vmin:g},vmax={p.vmax:g},seed={p.seed})"
                    )
            elif newparam.itype == ImageTypes.NORMALRANDOM:
                image.data = rng.normal(p.mu, p.sigma, size=shape)
                if image.title == DEFAULT_TITLE:
                    image.title = f"{prefix}(μ={p.mu:g},σ={p.sigma:g},seed={p.seed})"
            else:
                raise NotImplementedError(f"New param type: {newparam.itype.value}")
        return image
    return None



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Custom curve and image stats tool unit test.

PlotPy's statistics tools (`plotpy.tools.CurveStatsTool` and
`plotpy.tools.ImageStatsTool` are customized in `cdl.core.gui.docks` module).
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from __future__ import annotations

import numpy as np
import plotpy
from guidata.qthelpers import exec_dialog, qt_app_context
from plotpy.constants import PlotType
from plotpy.tests.unit.utils import drag_mouse
from plotpy.tools import CurveStatsTool, ImageStatsTool
from qtpy import QtWidgets as QW

import cdl.obj
from cdl.core.gui.docks import DataLabPlotWidget
from cdl.tests.data import create_multigauss_image, create_paracetamol_signal
from cdl.utils.misc import compare_versions


def simulate_stats_tool(
    plot_type: PlotType,
    obj: cdl.obj.SignalObj | cdl.obj.ImageObj,
    x_path: np.ndarray,
    y_path: np.ndarray,
) -> None:
    """Simulate stats tool with a custom signal or image."""
    widget = DataLabPlotWidget(plot_type)
    plot = widget.get_plot()
    item = obj.make_item()
    plot.add_item(item)
    plot.set_active_item(item)
    item.unselect()
    klass = CurveStatsTool if plot_type == PlotType.CURVE else ImageStatsTool
    stattool = widget.get_manager().get_tool(klass)
    stattool.activate()
    if compare_versions(plotpy.__version__, "<", "2.4"):
        qapp = QW.QApplication.instance()
        drag_mouse(widget, qapp, x_path, y_path)
    else:
        drag_mouse(widget, x_path, y_path)
    return widget


def test_stats_tool() -> None:
    """Test CurveStatsTool with a custom signal."""
    sig = create_paracetamol_signal()
    ima = create_multigauss_image()
    with qt_app_context():
        dlg = QW.QDialog()
        dlg.setWindowTitle("Stats Tool Test")
        dlg.resize(1200, 600)
        swidget = simulate_stats_tool(PlotType.CURVE, sig, [0.01, 0.02], [0.5, 0.5])
        iwidget = simulate_stats_tool(PlotType.IMAGE, ima, [0.2, 0.8], [0.7, 0.9])
        hlayout = QW.QHBoxLayout()
        hlayout.addWidget(swidget)
        hlayout.addWidget(iwidget)
        dlg.setLayout(hlayout)
    exec_dialog(dlg)


if __name__ == "__main__":
    test_stats_tool()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Coordinates Algorithms (see parent package :mod:`cdl.algorithms`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import numpy as np


def circle_to_diameter(
    xc: float, yc: float, r: float
) -> tuple[float, float, float, float]:
    """Convert circle center and radius to X diameter coordinates

    Args:
        xc: Circle center X coordinate
        yc: Circle center Y coordinate
        r: Circle radius

    Returns:
        tuple: Circle X diameter coordinates
    """
    return xc - r, yc, xc + r, yc


def array_circle_to_diameter(data: np.ndarray) -> np.ndarray:
    """Convert circle center and radius to X diameter coordinates (array version)

    Args:
        data: Circle center and radius, in the form of a 2D array (N, 3)

    Returns:
        Circle X diameter coordinates, in the form of a 2D array (N, 4)
    """
    xc, yc, r = data[:, 0], data[:, 1], data[:, 2]
    x_start = xc - r
    x_end = xc + r
    result = np.column_stack((x_start, yc, x_end, yc)).astype(float)
    return result


def circle_to_center_radius(
    x0: float, y0: float, x1: float, y1: float
) -> tuple[float, float, float]:
    """Convert circle X diameter coordinates to center and radius

    Args:
        x0: Diameter start X coordinate
        y0: Diameter start Y coordinate
        x1: Diameter end X coordinate
        y1: Diameter end Y coordinate

    Returns:
        tuple: Circle center and radius
    """
    xc, yc = (x0 + x1) / 2, (y0 + y1) / 2
    r = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2) / 2
    return xc, yc, r


def array_circle_to_center_radius(data: np.ndarray) -> np.ndarray:
    """Convert circle X diameter coordinates to center and radius (array version)

    Args:
        data: Circle X diameter coordinates, in the form of a 2D array (N, 4)

    Returns:
        Circle center and radius, in the form of a 2D array (N, 3)
    """
    x0, y0, x1, y1 = data[:, 0], data[:, 1], data[:, 2], data[:, 3]
    xc, yc = (x0 + x1) / 2, (y0 + y1) / 2
    r = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2) / 2
    result = np.column_stack((xc, yc, r)).astype(float)
    return result


def ellipse_to_diameters(
    xc: float, yc: float, a: float, b: float, theta: float
) -> tuple[float, float, float, float, float, float, float, float]:
    """Convert ellipse center, axes and angle to X/Y diameters coordinates

    Args:
        xc: Ellipse center X coordinate
        yc: Ellipse center Y coordinate
        a: Ellipse half larger axis
        b: Ellipse half smaller axis
        theta: Ellipse angle

    Returns:
        Ellipse X/Y diameters (major/minor axes) coordinates
    """
    dxa, dya = a * np.cos(theta), a * np.sin(theta)
    dxb, dyb = b * np.sin(theta), b * np.cos(theta)
    x0, y0, x1, y1 = xc - dxa, yc - dya, xc + dxa, yc + dya
    x2, y2, x3, y3 = xc - dxb, yc - dyb, xc + dxb, yc + dyb
    return x0, y0, x1, y1, x2, y2, x3, y3


def array_ellipse_to_diameters(data: np.ndarray) -> np.ndarray:
    """Convert ellipse center, axes and angle to X/Y diameters coordinates
    (array version)

    Args:
        data: Ellipse center, axes and angle, in the form of a 2D array (N, 5)

    Returns:
        Ellipse X/Y diameters (major/minor axes) coordinates,
         in the form of a 2D array (N, 8)
    """
    xc, yc, a, b, theta = data[:, 0], data[:, 1], data[:, 2], data[:, 3], data[:, 4]
    dxa, dya = a * np.cos(theta), a * np.sin(theta)
    dxb, dyb = b * np.sin(theta), b * np.cos(theta)
    x0, y0, x1, y1 = xc - dxa, yc - dya, xc + dxa, yc + dya
    x2, y2, x3, y3 = xc - dxb, yc - dyb, xc + dxb, yc + dyb
    result = np.column_stack((x0, y0, x1, y1, x2, y2, x3, y3)).astype(float)
    return result


def ellipse_to_center_axes_angle(
    x0: float,
    y0: float,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
) -> tuple[float, float, float, float, float]:
    """Convert ellipse X/Y diameters coordinates to center, axes and angle

    Args:
        x0: major axis start X coordinate
        y0: major axis start Y coordinate
        x1: major axis end X coordinate
        y1: major axis end Y coordinate
        x2: minor axis start X coordinate
        y2: minor axis start Y coordinate
        x3: minor axis end X coordinate
        y3: minor axis end Y coordinate

    Returns:
        Ellipse center, axes and angle
    """
    xc, yc = (x0 + x1) / 2, (y0 + y1) / 2
    a = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2) / 2
    b = np.sqrt((x3 - x2) ** 2 + (y3 - y2) ** 2) / 2
    theta = np.arctan2(y1 - y0, x1 - x0)
    return xc, yc, a, b, theta


def array_ellipse_to_center_axes_angle(data: np.ndarray) -> np.ndarray:
    """Convert ellipse X/Y diameters coordinates to center, axes and angle
    (array version)

    Args:
        data: Ellipse X/Y diameters coordinates, in the form of a 2D array (N, 8)

    Returns:
        Ellipse center, axes and angle, in the form of a 2D array (N, 5)
    """
    x0, y0, x1, y1, x2, y2, x3, y3 = (
        data[:, 0],
        data[:, 1],
        data[:, 2],
        data[:, 3],
        data[:, 4],
        data[:, 5],
        data[:, 6],
        data[:, 7],
    )
    xc, yc = (x0 + x1) / 2, (y0 + y1) / 2
    a = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2) / 2
    b = np.sqrt((x3 - x2) ** 2 + (y3 - y2) ** 2) / 2
    theta = np.arctan2(y1 - y0, x1 - x0)
    result = np.column_stack((xc, yc, a, b, theta)).astype(float)
    return result



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""Make a WiX Toolset .wxs file for the DataLab Windows installer.

History:

- Version 1.0.0 (2024-04-01)
    Initial version.
"""

# TODO: [P4] Localization, eventually.

import argparse
import os
import os.path as osp
import re
import uuid
import xml.etree.ElementTree as ET

COUNT = 0


def generate_id() -> str:
    """Generate an ID for a WiX Toolset XML element."""
    global COUNT
    COUNT += 1
    return f"ID_{COUNT:04d}"


def insert_text_after(text: str, containing: str, content: str) -> str:
    """Insert line of text after the line containing a specific text."""
    if os.linesep in content:
        linesep = os.linesep
    elif "\r\n" in content:
        linesep = "\r\n"
    else:
        linesep = "\n"
    lines = content.splitlines()
    for i_line, line in enumerate(lines):
        if containing in line:
            lines.insert(i_line + 1, text)
            break
    return linesep.join(lines)


def make_wxs(product_name: str, version: str) -> None:
    """Make a .wxs file for the DataLab Windows installer."""
    # MSI Version does not support labels, at least not in a way compatible with
    # Python packaging (e.g., "0.16.dev0" is not a valid MSI version).
    #
    # Here is the conversion we must do:
    # - "0.16.post1" -> "0.16.1"
    # - "0.16.post2" -> "0.16.1"
    # - "0.16.dev0" -> "0.16.0"
    # - "0.16.dev1" -> "0.16.0"
    # - "0.16.alpha1" -> "0.16.0"
    # - "0.16.alpha2" -> "0.16.0"
    # - "0.16.beta1" -> "0.16.0"
    # - "0.16.beta2" -> "0.16.0"
    # - "0.16.rc1" -> "0.16.0"
    # - "0.16.rc2" -> "0.16.0"
    #
    version_msi = re.sub(r"\.post\d+$", ".1", version)
    version_msi = re.sub(r"\.dev\d+$", ".0", version_msi)
    version_msi = re.sub(r"(rc|a|b)\d+$", ".0", version_msi)

    wix_dir = osp.abspath(osp.dirname(__file__))
    proj_dir = osp.join(wix_dir, os.pardir)
    dist_dir = osp.join(proj_dir, "dist", product_name)
    wxs_path = osp.join(wix_dir, f"generic-{product_name}.wxs")
    output_path = osp.join(wix_dir, f"{product_name}-{version}.wxs")

    dir_ids: dict[str, str] = {}
    file_ids: dict[str, str] = {}

    files_dict: dict[str, list[str]] = {}

    dir_str_list: list[str] = []
    comp_str_list: list[str] = []

    for pth in os.listdir(dist_dir):
        root_dir = osp.join(dist_dir, pth)

        if not osp.isdir(root_dir):
            continue

        dpath_list: list[str] = []
        for root, dirs, filenames in os.walk(root_dir):
            for dpath in dirs:
                relpath = osp.relpath(osp.join(root, dpath), proj_dir)
                dpath_list.append(relpath)
                dir_ids[relpath] = generate_id()
                files_dict.setdefault(osp.dirname(relpath), [])
            for filename in filenames:
                relpath = osp.relpath(osp.join(root, filename), proj_dir)
                file_ids[relpath] = generate_id()
                files_dict.setdefault(osp.dirname(relpath), []).append(relpath)

        # Create the base directory structure in XML:
        base_name = osp.basename(root_dir)
        base_path = osp.relpath(root_dir, proj_dir)
        base_id = dir_ids[base_path] = generate_id()
        dir_xml = ET.Element("Directory", Id=base_id, Name=base_name)

        # Nesting directories, recursively, in XML:
        for dpath in sorted(dpath_list):
            dname = osp.basename(dpath)
            parent = dir_xml
            for element in parent.iter():
                if element.get("Id") == dir_ids[osp.dirname(dpath)]:
                    parent = element
                    break
            else:
                raise ValueError(f"Parent directory not found for {dpath}")
            ET.SubElement(parent, "Directory", Id=dir_ids[dpath], Name=dname)
        space = " " * 4
        ET.indent(dir_xml, space=space, level=4)
        dir_str_list.append(space * 4 + ET.tostring(dir_xml, encoding="unicode"))

        # Create additionnal components for each file in the directory structure:
        for dpath in sorted([base_path] + dpath_list):
            did = dir_ids[dpath]
            files = files_dict.get(dpath, [])
            if files:
                # This is a directory with files, so we need to create components:
                for path in files:
                    fid = file_ids[path]
                    guid = str(uuid.uuid4())
                    comp_xml = ET.Element("Component", Id=fid, Directory=did, Guid=guid)
                    ET.SubElement(comp_xml, "File", Source=path, KeyPath="yes")
                    ET.indent(comp_xml, space=space, level=3)
                    comp_str = space * 3 + ET.tostring(comp_xml, encoding="unicode")
                    comp_str_list.append(comp_str)
            elif dpath != base_path:
                # This is an empty directory, so we need to create a folder:
                guid = str(uuid.uuid4())
                cdid = f"CreateFolder_{did}"
                comp_xml = ET.Element("Component", Id=cdid, Directory=did, Guid=guid)
                ET.SubElement(comp_xml, "CreateFolder")
                ET.indent(comp_xml, space=space, level=3)
                comp_str = space * 3 + ET.tostring(comp_xml, encoding="unicode")
                comp_str_list.append(comp_str)

    dir_str = "\n".join(dir_str_list).replace("><", ">\n<")
    # print("Directory structure:\n", dir_str)
    comp_str = "\n".join(comp_str_list).replace("><", ">\n<")
    # print("Component structure:\n", comp_str)

    # Create the .wxs file:
    with open(wxs_path, "r", encoding="utf-8") as fd:
        wxs = fd.read()
    wxs = insert_text_after(dir_str, "<!-- Automatically inserted directories -->", wxs)
    wxs = insert_text_after(comp_str, "<!-- Automatically inserted components -->", wxs)
    wxs = wxs.replace("{version}", version_msi)
    with open(output_path, "w", encoding="utf-8") as fd:
        fd.write(wxs)
    print("Successfully created:", output_path)


def run() -> None:
    """Run the script."""
    parser = argparse.ArgumentParser(
        description="Make a WiX Toolset .wxs file for the DataLab Windows installer."
    )
    parser.add_argument("product_name", help="Product name")
    parser.add_argument("version", help="Product version")
    args = parser.parse_args()
    make_wxs(args.product_name, args.version)


if __name__ == "__main__":
    if len(os.sys.argv) == 1:
        # For testing purposes:
        make_wxs("DataLab", "0.14.2")
    else:
        run()

    # After making the .wxs file, run the following command to create the .msi file:
    #   wix build .\wix\DataLab.wxs -ext WixToolset.UI.wixext
